# Copyright 2012, Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging
from unittest import mock

import fixtures
from oslotest import base as test_base

from oslo_utils import excutils
from oslo_utils import timeutils


class Fail1(excutils.CausedByException):
    pass


class Fail2(excutils.CausedByException):
    pass


class CausedByTest(test_base.BaseTestCase):
    def test_caused_by_explicit(self):
        e = self.assertRaises(
            Fail1,
            excutils.raise_with_cause,
            Fail1,
            "I was broken",
            cause=Fail2("I have been broken"),
        )
        self.assertIsInstance(e.cause, Fail2)
        e_p = e.pformat()
        self.assertIn("I have been broken", e_p)
        self.assertIn("Fail2", e_p)

    def test_caused_by_implicit(self):
        def raises_chained():
            try:
                raise Fail2("I have been broken")
            except Fail2:
                excutils.raise_with_cause(Fail1, "I was broken")

        e = self.assertRaises(Fail1, raises_chained)
        self.assertIsInstance(e.cause, Fail2)
        e_p = e.pformat()
        self.assertIn("I have been broken", e_p)
        self.assertIn("Fail2", e_p)


class SaveAndReraiseTest(test_base.BaseTestCase):
    def test_save_and_reraise_exception_forced(self):
        def _force_reraise():
            try:
                raise OSError("I broke")
            except Exception:
                with excutils.save_and_reraise_exception() as e:
                    e.reraise = False
                e.force_reraise()

        self.assertRaises(IOError, _force_reraise)

    def test_save_and_reraise_exception_capture_reraise(self):
        def _force_reraise():
            try:
                raise OSError("I broke")
            except Exception:
                excutils.save_and_reraise_exception().capture().force_reraise()

        self.assertRaises(IOError, _force_reraise)

    def test_save_and_reraise_exception_capture_not_active(self):
        e = excutils.save_and_reraise_exception()
        self.assertRaises(RuntimeError, e.capture, check=True)

    def test_save_and_reraise_exception_forced_not_active(self):
        e = excutils.save_and_reraise_exception()
        self.assertRaises(RuntimeError, e.force_reraise)
        e = excutils.save_and_reraise_exception()
        e.capture(check=False)
        self.assertRaises(RuntimeError, e.force_reraise)

    def test_save_and_reraise_exception(self):
        e = None
        msg = 'foo'
        try:
            try:
                raise Exception(msg)
            except Exception:
                with excutils.save_and_reraise_exception():
                    pass
        except Exception as _e:
            e = _e

        self.assertEqual(str(e), msg)

    @mock.patch('logging.getLogger')
    def test_save_and_reraise_exception_dropped(self, get_logger_mock):
        logger = get_logger_mock()
        e = None
        msg = 'second exception'
        try:
            try:
                raise Exception('dropped')
            except Exception:
                with excutils.save_and_reraise_exception():
                    raise Exception(msg)
        except Exception as _e:
            e = _e
        self.assertEqual(str(e), msg)
        self.assertTrue(logger.error.called)

    def test_save_and_reraise_exception_no_reraise(self):
        """Test that suppressing the reraise works."""
        try:
            raise Exception('foo')
        except Exception:
            with excutils.save_and_reraise_exception() as ctxt:
                ctxt.reraise = False

    @mock.patch('logging.getLogger')
    def test_save_and_reraise_exception_dropped_no_reraise(
        self, get_logger_mock
    ):
        logger = get_logger_mock()
        e = None
        msg = 'second exception'
        try:
            try:
                raise Exception('dropped')
            except Exception:
                with excutils.save_and_reraise_exception(reraise=False):
                    raise Exception(msg)
        except Exception as _e:
            e = _e
        self.assertEqual(str(e), msg)
        self.assertFalse(logger.error.called)

    def test_save_and_reraise_exception_provided_logger(self):
        fake_logger = mock.MagicMock()
        try:
            try:
                raise Exception('foo')
            except Exception:
                with excutils.save_and_reraise_exception(logger=fake_logger):
                    raise Exception('second exception')
        except Exception:
            pass
        self.assertTrue(fake_logger.error.called)


class ForeverRetryUncaughtExceptionsTest(test_base.BaseTestCase):
    def setUp(self):
        super().setUp()

        self._exceptions = []

        self.useFixture(fixtures.MockPatch('time.sleep', return_value=None))

    @excutils.forever_retry_uncaught_exceptions
    def exception_generator(self):
        while self._exceptions:
            raise self._exceptions.pop(0)

    @mock.patch.object(logging, 'exception')
    @mock.patch.object(timeutils, 'now')
    def test_exc_retrier_1exc_gives_1log(self, mock_now, mock_log):
        self._exceptions = [Exception('unexpected 1')]
        mock_now.side_effect = [0]

        self.exception_generator()

        self.assertEqual([], self._exceptions)
        # log should only be called once
        mock_log.assert_called_once_with(
            'Unexpected exception occurred 1 time(s)... retrying.'
        )
        mock_now.assert_has_calls(
            [
                mock.call(),
            ]
        )

    @mock.patch.object(logging, 'exception')
    @mock.patch.object(timeutils, 'now')
    def test_exc_retrier_same_10exc_1min_gives_1log(self, mock_now, mock_log):
        self._exceptions = [
            Exception('unexpected 1'),
        ]
        # Timestamp calls that happen after the logging is possibly triggered.
        mock_now_side_effect = [0]

        # By design, the following exceptions won't get logged because they
        # are within the same minute.
        for i in range(2, 11):
            self._exceptions.append(Exception('unexpected 1'))
            # Timestamp calls that happen before the logging is possibly
            # triggered.
            mock_now_side_effect.append(i)

        mock_now.side_effect = mock_now_side_effect

        self.exception_generator()

        self.assertEqual([], self._exceptions)
        self.assertEqual(10, len(mock_now.mock_calls))
        self.assertEqual(1, len(mock_log.mock_calls))
        mock_log.assert_has_calls(
            [
                mock.call(
                    'Unexpected exception occurred 1 time(s)... retrying.'
                ),
            ]
        )

    @mock.patch.object(logging, 'exception')
    @mock.patch.object(timeutils, 'now')
    def test_exc_retrier_same_2exc_2min_gives_2logs(self, mock_now, mock_log):
        self._exceptions = [
            Exception('unexpected 1'),
            Exception('unexpected 1'),
        ]

        mock_now.side_effect = [
            # Timestamp calls that happen after the logging is possibly
            # triggered
            0,
            # Timestamp calls that happen before the logging is possibly
            # triggered
            65,
            # Timestamp calls that happen after the logging is possibly
            # triggered.
            65,
            66,
        ]

        self.exception_generator()

        self.assertEqual([], self._exceptions)
        self.assertEqual(4, len(mock_now.mock_calls))
        self.assertEqual(2, len(mock_log.mock_calls))
        mock_log.assert_has_calls(
            [
                mock.call(
                    'Unexpected exception occurred 1 time(s)... retrying.'
                ),
                mock.call(
                    'Unexpected exception occurred 1 time(s)... retrying.'
                ),
            ]
        )

    @mock.patch.object(logging, 'exception')
    @mock.patch.object(timeutils, 'now')
    def test_exc_retrier_same_10exc_2min_gives_2logs(self, mock_now, mock_log):
        self._exceptions = [
            Exception('unexpected 1'),
        ]

        # Timestamp calls that happen after the logging is possibly triggered.
        mock_now_side_effect = [
            0,
        ]

        for ts in [12, 23, 34, 45]:
            self._exceptions.append(Exception('unexpected 1'))
            # Timestamp calls that happen before the logging is possibly
            # triggered.
            mock_now_side_effect.append(ts)

        # The previous 4 exceptions are counted here

        self._exceptions.append(Exception('unexpected 1'))
        # Timestamp calls that happen before the logging is possibly triggered.
        mock_now_side_effect.append(106)

        for ts in [106, 107]:
            # Timestamp calls that happen after the logging is possibly
            # triggered.
            mock_now_side_effect.append(ts)

        # Again, the following are not logged due to being within
        # the same minute

        for ts in [117, 128, 139, 150]:
            self._exceptions.append(Exception('unexpected 1'))
            # Timestamp calls that happen before the logging is possibly
            # triggered.
            mock_now_side_effect.append(ts)

        mock_now.side_effect = mock_now_side_effect

        self.exception_generator()

        self.assertEqual([], self._exceptions)
        self.assertEqual(12, len(mock_now.mock_calls))
        self.assertEqual(2, len(mock_log.mock_calls))
        mock_log.assert_has_calls(
            [
                mock.call(
                    'Unexpected exception occurred 1 time(s)... retrying.'
                ),
                mock.call(
                    'Unexpected exception occurred 5 time(s)... retrying.'
                ),
            ]
        )

    @mock.patch.object(logging, 'exception')
    @mock.patch.object(timeutils, 'now')
    def test_exc_retrier_mixed_4exc_1min_gives_2logs(self, mock_now, mock_log):
        # The stop watch will be started, which will consume one timestamp
        # call.

        self._exceptions = [
            Exception('unexpected 1'),
        ]
        # Timestamp calls that happen after the logging is possibly
        # triggered.
        mock_now_side_effect = [0]

        # By design, this second 'unexpected 1' exception is not counted.  This
        # is likely a rare thing and is a sacrifice for code simplicity.
        self._exceptions.append(Exception('unexpected 1'))

        # Timestamp calls that happen before the logging is possibly triggered.
        # Since the exception will be the same the expiry method will be
        # called, which uses up a timestamp call.
        mock_now_side_effect.append(5)

        self._exceptions.append(Exception('unexpected 2'))
        # Timestamp calls that happen after the logging is possibly triggered.
        # The watch should get reset, which uses up two timestamp calls.
        mock_now_side_effect.extend([10, 20])

        self._exceptions.append(Exception('unexpected 2'))
        # Timestamp calls that happen before the logging is possibly triggered.
        # Since the exception will be the same the expiry method will be
        # called, which uses up a timestamp call.
        mock_now_side_effect.append(25)

        mock_now.side_effect = mock_now_side_effect

        self.exception_generator()

        self.assertEqual([], self._exceptions)
        self.assertEqual(5, len(mock_now.mock_calls))
        self.assertEqual(2, len(mock_log.mock_calls))
        mock_log.assert_has_calls(
            [
                mock.call(
                    'Unexpected exception occurred 1 time(s)... retrying.'
                ),
                mock.call(
                    'Unexpected exception occurred 1 time(s)... retrying.'
                ),
            ]
        )

    @mock.patch.object(logging, 'exception')
    @mock.patch.object(timeutils, 'now')
    def test_exc_retrier_mixed_4exc_2min_gives_2logs(self, mock_now, mock_log):
        self._exceptions = [
            Exception('unexpected 1'),
        ]
        # Timestamp calls that happen after the logging is possibly triggered.
        mock_now_side_effect = [0]

        # Again, this second exception of the same type is not counted
        # for the sake of code simplicity.

        self._exceptions.append(Exception('unexpected 1'))
        # Timestamp calls that happen before the logging is possibly triggered.
        mock_now_side_effect.append(10)

        # The difference between this and the previous case is the log
        # is also triggered by more than a minute expiring.

        self._exceptions.append(Exception('unexpected 2'))
        # Timestamp calls that happen after the logging is possibly triggered.
        mock_now_side_effect.extend([100, 105])

        self._exceptions.append(Exception('unexpected 2'))
        # Timestamp calls that happen before the logging is possibly triggered.
        mock_now_side_effect.append(110)

        mock_now.side_effect = mock_now_side_effect

        self.exception_generator()

        self.assertEqual([], self._exceptions)
        self.assertEqual(5, len(mock_now.mock_calls))
        self.assertEqual(2, len(mock_log.mock_calls))
        mock_log.assert_has_calls(
            [
                mock.call(
                    'Unexpected exception occurred 1 time(s)... retrying.'
                ),
                mock.call(
                    'Unexpected exception occurred 1 time(s)... retrying.'
                ),
            ]
        )

    @mock.patch.object(logging, 'exception')
    @mock.patch.object(timeutils, 'now')
    def test_exc_retrier_mixed_4exc_2min_gives_3logs(self, mock_now, mock_log):
        self._exceptions = [
            Exception('unexpected 1'),
        ]
        # Timestamp calls that happen after the logging is possibly triggered.
        mock_now_side_effect = [0]

        # This time the second 'unexpected 1' exception is counted due
        # to the same exception occurring same when the minute expires.

        self._exceptions.append(Exception('unexpected 1'))
        # Timestamp calls that happen before the logging is possibly triggered.
        mock_now_side_effect.append(10)

        self._exceptions.append(Exception('unexpected 1'))
        # Timestamp calls that happen before the logging is possibly triggered.
        mock_now_side_effect.extend([100, 100, 105])

        self._exceptions.append(Exception('unexpected 2'))
        # Timestamp calls that happen after the logging is possibly triggered.
        mock_now_side_effect.extend([110, 111])

        mock_now.side_effect = mock_now_side_effect

        self.exception_generator()

        self.assertEqual([], self._exceptions)
        self.assertEqual(7, len(mock_now.mock_calls))
        self.assertEqual(3, len(mock_log.mock_calls))
        mock_log.assert_has_calls(
            [
                mock.call(
                    'Unexpected exception occurred 1 time(s)... retrying.'
                ),
                mock.call(
                    'Unexpected exception occurred 2 time(s)... retrying.'
                ),
                mock.call(
                    'Unexpected exception occurred 1 time(s)... retrying.'
                ),
            ]
        )


class ExceptionFilterTest(test_base.BaseTestCase):
    def _make_filter_func(self, ignore_classes=AssertionError):
        @excutils.exception_filter
        def ignore_exceptions(ex):
            """Ignore some exceptions F."""
            return isinstance(ex, ignore_classes)

        return ignore_exceptions

    def _make_filter_method(self, ignore_classes=AssertionError):
        class ExceptionIgnorer:
            def __init__(self, ignore):
                self.ignore = ignore

            @excutils.exception_filter  # type: ignore
            def ignore_exceptions(self, ex):
                """Ignore some exceptions M."""
                return isinstance(ex, self.ignore)

        return ExceptionIgnorer(ignore_classes).ignore_exceptions

    def _make_filter_classmethod(self, ignore_classes=AssertionError):
        class ExceptionIgnorer:
            ignore = ignore_classes

            @excutils.exception_filter  # type: ignore
            @classmethod
            def ignore_exceptions(cls, ex):
                """Ignore some exceptions C."""
                return isinstance(ex, cls.ignore)

        return ExceptionIgnorer.ignore_exceptions

    def _make_filter_staticmethod(self, ignore_classes=AssertionError):
        class ExceptionIgnorer:
            @excutils.exception_filter
            @staticmethod
            def ignore_exceptions(ex):
                """Ignore some exceptions S."""
                return isinstance(ex, ignore_classes)

        return ExceptionIgnorer.ignore_exceptions

    def test_filter_func_call(self):
        ignore_assertion_error = self._make_filter_func()

        try:
            assert False, "This is a test"
        except Exception as exc:
            ignore_assertion_error(exc)

    def test_raise_func_call(self):
        ignore_assertion_error = self._make_filter_func()

        try:
            raise RuntimeError
        except Exception as exc:
            self.assertRaises(RuntimeError, ignore_assertion_error, exc)

    def test_raise_previous_func_call(self):
        ignore_assertion_error = self._make_filter_func()

        try:
            raise RuntimeError
        except Exception as exc1:
            try:
                raise RuntimeError
            except Exception as exc2:
                self.assertIsNot(exc1, exc2)
            raised = self.assertRaises(
                RuntimeError, ignore_assertion_error, exc1
            )
            self.assertIs(exc1, raised)

    def test_raise_previous_after_filtered_func_call(self):
        ignore_assertion_error = self._make_filter_func()

        try:
            raise RuntimeError
        except Exception as exc1:
            try:
                assert False, "This is a test"
            except Exception:
                pass
            self.assertRaises(RuntimeError, ignore_assertion_error, exc1)

    def test_raise_other_func_call(self):
        @excutils.exception_filter
        def translate_exceptions(ex):
            raise RuntimeError

        try:
            assert False, "This is a test"
        except Exception as exc:
            self.assertRaises(RuntimeError, translate_exceptions, exc)

    def test_filter_func_context_manager(self):
        ignore_assertion_error = self._make_filter_func()

        with ignore_assertion_error:
            assert False, "This is a test"

    def test_raise_func_context_manager(self):
        ignore_assertion_error = self._make_filter_func()

        def try_runtime_err():
            with ignore_assertion_error:
                raise RuntimeError

        self.assertRaises(RuntimeError, try_runtime_err)

    def test_raise_other_func_context_manager(self):
        @excutils.exception_filter
        def translate_exceptions(ex):
            raise RuntimeError

        def try_assertion():
            with translate_exceptions:
                assert False, "This is a test"

        self.assertRaises(RuntimeError, try_assertion)

    def test_noexc_func_context_manager(self):
        ignore_assertion_error = self._make_filter_func()

        with ignore_assertion_error:
            pass

    def test_noexc_nocall_func_context_manager(self):
        @excutils.exception_filter
        def translate_exceptions(ex):
            raise RuntimeError

        with translate_exceptions:
            pass

    def test_func_docstring(self):
        ignore_func = self._make_filter_func()
        self.assertEqual('Ignore some exceptions F.', ignore_func.__doc__)

    def test_filter_method_call(self):
        ignore_assertion_error = self._make_filter_method()

        try:
            assert False, "This is a test"
        except Exception as exc:
            ignore_assertion_error(exc)

    def test_raise_method_call(self):
        ignore_assertion_error = self._make_filter_method()

        try:
            raise RuntimeError
        except Exception as exc:
            self.assertRaises(RuntimeError, ignore_assertion_error, exc)

    def test_filter_method_context_manager(self):
        ignore_assertion_error = self._make_filter_method()

        with ignore_assertion_error:
            assert False, "This is a test"

    def test_raise_method_context_manager(self):
        ignore_assertion_error = self._make_filter_method()

        def try_runtime_err():
            with ignore_assertion_error:
                raise RuntimeError

        self.assertRaises(RuntimeError, try_runtime_err)

    def test_method_docstring(self):
        ignore_func = self._make_filter_method()
        self.assertEqual('Ignore some exceptions M.', ignore_func.__doc__)

    def test_filter_classmethod_call(self):
        ignore_assertion_error = self._make_filter_classmethod()

        try:
            assert False, "This is a test"
        except Exception as exc:
            ignore_assertion_error(exc)

    def test_raise_classmethod_call(self):
        ignore_assertion_error = self._make_filter_classmethod()

        try:
            raise RuntimeError
        except Exception as exc:
            self.assertRaises(RuntimeError, ignore_assertion_error, exc)

    def test_filter_classmethod_context_manager(self):
        ignore_assertion_error = self._make_filter_classmethod()

        with ignore_assertion_error:
            assert False, "This is a test"

    def test_raise_classmethod_context_manager(self):
        ignore_assertion_error = self._make_filter_classmethod()

        def try_runtime_err():
            with ignore_assertion_error:
                raise RuntimeError

        self.assertRaises(RuntimeError, try_runtime_err)

    def test_classmethod_docstring(self):
        ignore_func = self._make_filter_classmethod()
        self.assertEqual('Ignore some exceptions C.', ignore_func.__doc__)

    def test_filter_staticmethod_call(self):
        ignore_assertion_error = self._make_filter_staticmethod()

        try:
            assert False, "This is a test"
        except Exception as exc:
            ignore_assertion_error(exc)

    def test_raise_staticmethod_call(self):
        ignore_assertion_error = self._make_filter_staticmethod()

        try:
            raise RuntimeError
        except Exception as exc:
            self.assertRaises(RuntimeError, ignore_assertion_error, exc)

    def test_filter_staticmethod_context_manager(self):
        ignore_assertion_error = self._make_filter_staticmethod()

        with ignore_assertion_error:
            assert False, "This is a test"

    def test_raise_staticmethod_context_manager(self):
        ignore_assertion_error = self._make_filter_staticmethod()

        def try_runtime_err():
            with ignore_assertion_error:
                raise RuntimeError

        self.assertRaises(RuntimeError, try_runtime_err)

    def test_staticmethod_docstring(self):
        ignore_func = self._make_filter_staticmethod()
        self.assertEqual('Ignore some exceptions S.', ignore_func.__doc__)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslotest import base as test_base

from oslo_utils import specs_matcher


class SpecsMatcherTestCase(test_base.BaseTestCase):
    def _do_specs_matcher_test(self, value, req, matches):
        assertion = self.assertTrue if matches else self.assertFalse
        assertion(specs_matcher.match(value, req))

    def test_specs_matches_simple(self):
        self._do_specs_matcher_test(value='1', req='1', matches=True)

    def test_specs_fails_string_vs_int(self):
        # With no operator specified it is a string comparison test, therefore
        # '1' does not equal '01'
        self._do_specs_matcher_test(value='01', req='1', matches=False)

    def test_specs_match_int_leading_zero(self):
        # Show that numerical comparison works with leading zero
        self._do_specs_matcher_test(value='01', req='== 1', matches=True)

    def test_specs_fails_simple(self):
        self._do_specs_matcher_test(value='', req='1', matches=False)

    def test_specs_fails_simple2(self):
        self._do_specs_matcher_test(value='3', req='1', matches=False)

    def test_specs_fails_simple3(self):
        self._do_specs_matcher_test(value='222', req='2', matches=False)

    def test_specs_fails_with_bogus_ops(self):
        self._do_specs_matcher_test(value='4', req='! 2', matches=False)

    def test_specs_matches_with_op_eq(self):
        self._do_specs_matcher_test(value='123', req='= 123', matches=True)

    def test_specs_matches_with_op_eq2(self):
        self._do_specs_matcher_test(value='124', req='= 123', matches=True)

    def test_specs_fails_with_op_eq(self):
        self._do_specs_matcher_test(value='34', req='= 234', matches=False)

    def test_specs_fails_with_op_eq3(self):
        self._do_specs_matcher_test(value='34', req='=', matches=False)

    def test_specs_matches_with_op_seq(self):
        self._do_specs_matcher_test(value='123', req='s== 123', matches=True)

    def test_specs_fails_with_op_seq(self):
        self._do_specs_matcher_test(value='1234', req='s== 123', matches=False)

    def test_specs_matches_with_op_sneq(self):
        self._do_specs_matcher_test(value='1234', req='s!= 123', matches=True)

    def test_specs_fails_with_op_sneq(self):
        self._do_specs_matcher_test(value='123', req='s!= 123', matches=False)

    def test_specs_matches_with_op_sge(self):
        self._do_specs_matcher_test(value='234', req='s>= 1000', matches=True)

    def test_specs_matches_with_op_sge2(self):
        self._do_specs_matcher_test(value='234', req='s>= 234', matches=True)

    def test_specs_fails_with_op_sge(self):
        self._do_specs_matcher_test(value='1000', req='s>= 234', matches=False)

    def test_specs_matches_with_op_sle(self):
        self._do_specs_matcher_test(value='1000', req='s<= 1234', matches=True)

    def test_specs_matches_with_op_sle2(self):
        self._do_specs_matcher_test(value='1234', req='s<= 1234', matches=True)

    def test_specs_fails_with_op_sle(self):
        self._do_specs_matcher_test(
            value='1234', req='s<= 1000', matches=False
        )

    def test_specs_matches_with_op_sl(self):
        self._do_specs_matcher_test(value='12', req='s< 2', matches=True)

    def test_specs_fails_with_op_sl(self):
        self._do_specs_matcher_test(value='2', req='s< 12', matches=False)

    def test_specs_fails_with_op_sl2(self):
        self._do_specs_matcher_test(value='12', req='s< 12', matches=False)

    def test_specs_matches_with_op_sg(self):
        self._do_specs_matcher_test(value='2', req='s> 12', matches=True)

    def test_specs_fails_with_op_sg(self):
        self._do_specs_matcher_test(value='12', req='s> 2', matches=False)

    def test_specs_fails_with_op_sg2(self):
        self._do_specs_matcher_test(value='12', req='s> 12', matches=False)

    def test_specs_matches_with_op_in(self):
        self._do_specs_matcher_test(
            value='12311321', req='<in> 11', matches=True
        )

    def test_specs_matches_with_op_in2(self):
        self._do_specs_matcher_test(
            value='12311321', req='<in> 12311321', matches=True
        )

    def test_specs_matches_with_op_in3(self):
        self._do_specs_matcher_test(
            value='12311321', req='<in> 12311321 <in>', matches=True
        )

    def test_specs_fails_with_op_in(self):
        self._do_specs_matcher_test(
            value='12310321', req='<in> 11', matches=False
        )

    def test_specs_fails_with_op_in2(self):
        self._do_specs_matcher_test(
            value='12310321', req='<in> 11 <in>', matches=False
        )

    def test_specs_matches_with_op_or(self):
        self._do_specs_matcher_test(
            value='12', req='<or> 11 <or> 12', matches=True
        )

    def test_specs_matches_with_op_or2(self):
        self._do_specs_matcher_test(
            value='12', req='<or> 11 <or> 12 <or>', matches=True
        )

    def test_specs_matches_with_op_or3(self):
        self._do_specs_matcher_test(value='12', req='<or> 12', matches=True)

    def test_specs_fails_with_op_or(self):
        self._do_specs_matcher_test(
            value='13', req='<or> 11 <or> 12', matches=False
        )

    def test_specs_fails_with_op_or2(self):
        self._do_specs_matcher_test(
            value='13', req='<or> 11 <or> 12 <or>', matches=False
        )

    def test_specs_fails_with_op_or3(self):
        self._do_specs_matcher_test(value='13', req='<or> 11', matches=False)

    def test_specs_matches_with_op_le(self):
        self._do_specs_matcher_test(value='2', req='<= 10', matches=True)

    def test_specs_matches_with_op_le2(self):
        self._do_specs_matcher_test(value='10', req='<= 10', matches=True)

    def test_specs_fails_with_op_le(self):
        self._do_specs_matcher_test(value='3', req='<= 2', matches=False)

    def test_specs_matches_with_op_ge(self):
        self._do_specs_matcher_test(value='3', req='>= 1', matches=True)

    def test_specs_matches_with_op_ge2(self):
        self._do_specs_matcher_test(value='3.0', req='>= 3', matches=True)

    def test_specs_matches_with_op_g(self):
        self._do_specs_matcher_test(value='3', req='> 1', matches=True)

    def test_specs_matches_with_op_g2(self):
        self._do_specs_matcher_test(value='3', req='> 3', matches=False)

    def test_specs_matches_with_op_g3(self):
        self._do_specs_matcher_test(value='3.0', req='> 2', matches=True)

    def test_specs_matches_with_op_l(self):
        self._do_specs_matcher_test(value='3', req='< 5', matches=True)

    def test_specs_matches_with_op_l2(self):
        self._do_specs_matcher_test(value='3', req='< 3', matches=False)

    def test_specs_matches_with_op_l3(self):
        self._do_specs_matcher_test(value='1.0', req='< 6', matches=True)

    def test_specs_fails_with_op_ge(self):
        self._do_specs_matcher_test(value='2', req='>= 3', matches=False)

    def test_specs_matches_with_op_ne(self):
        self._do_specs_matcher_test(value='3.2', req='!= 3.1', matches=True)

    def test_specs_fails_with_op_ne(self):
        self._do_specs_matcher_test(value='3.2', req='!= 3.2', matches=False)

    def test_specs_matches_with_op_eqeq(self):
        self._do_specs_matcher_test(value='3', req='== 3', matches=True)

    def test_specs_matches_with_op_eqeq2(self):
        self._do_specs_matcher_test(value='3.0', req='== 3', matches=True)

    def test_specs_fails_with_op_eqeq(self):
        self._do_specs_matcher_test(value='3.0', req='== 3.1', matches=False)

    def test_specs_matches_all_with_op_allin(self):
        self._do_specs_matcher_test(
            value=str(['aes', 'mmx', 'aux']),
            req='<all-in> aes mmx',
            matches=True,
        )

    def test_specs_matches_one_with_op_allin(self):
        self._do_specs_matcher_test(
            value=str(['aes', 'mmx', 'aux']), req='<all-in> mmx', matches=True
        )

    def test_specs_fails_with_op_allin(self):
        self._do_specs_matcher_test(
            value=str(['aes', 'mmx', 'aux']),
            req='<all-in>  txt',
            matches=False,
        )

    def test_specs_fails_all_with_op_allin(self):
        self._do_specs_matcher_test(
            value=str(['aes', 'mmx', 'aux']),
            req='<all-in> txt 3dnow',
            matches=False,
        )

    def test_specs_fails_match_one_with_op_allin(self):
        self._do_specs_matcher_test(
            value=str(['aes', 'mmx', 'aux']),
            req='<all-in> txt aes',
            matches=False,
        )

    def test_specs_fails_match_substr_single(self):
        self._do_specs_matcher_test(
            value=str(['X_X']), req='<all-in> _', matches=False
        )

    def test_specs_fails_match_substr(self):
        self._do_specs_matcher_test(
            value=str(['X___X']), req='<all-in> ___', matches=False
        )

    def test_specs_fails_match_substr_reversed(self):
        self._do_specs_matcher_test(
            value=str(['aes', 'mmx', 'aux']),
            req='<all-in> XaesX',
            matches=False,
        )

    def test_specs_fails_onechar_with_op_allin(self):
        self.assertRaises(
            TypeError,
            specs_matcher.match,
            value=str(['aes', 'mmx', 'aux']),
            req='<all-in> e',
        )

    def test_specs_errors_list_with_op_allin(self):
        self.assertRaises(
            TypeError,
            specs_matcher.match,
            value=['aes', 'mmx', 'aux'],
            req='<all-in> aes',
        )

    def test_specs_errors_str_with_op_allin(self):
        self.assertRaises(
            TypeError, specs_matcher.match, value='aes', req='<all-in> aes'
        )

    def test_specs_errors_dict_literal_with_op_allin(self):
        self.assertRaises(
            TypeError,
            specs_matcher.match,
            value=str({'aes': 1}),
            req='<all-in> aes',
        )

    def test_specs_errors_bad_literal_with_op_allin(self):
        self.assertRaises(
            TypeError, specs_matcher.match, value="^&*($", req='<all-in> aes'
        )

    def test_specs_fails_not_enough_args_with_op_rangein(self):
        self.assertRaises(
            TypeError, specs_matcher.match, value="23", req='<range-in> [ 10 ]'
        )

    def test_specs_fails_no_brackets_with_op_rangein(self):
        self.assertRaises(
            TypeError, specs_matcher.match, value="23", req='<range-in> 10 20'
        )

    def test_specs_fails_no_opening_bracket_with_op_rangein(self):
        self.assertRaises(
            TypeError,
            specs_matcher.match,
            value="23",
            req='<range-in> 10 20 ]',
        )

    def test_specs_fails_no_closing_bracket_with_op_rangein(self):
        self.assertRaises(
            TypeError,
            specs_matcher.match,
            value="23",
            req='<range-in> [ 10 20',
        )

    def test_specs_fails_invalid_brackets_with_op_rangein(self):
        self.assertRaises(
            TypeError,
            specs_matcher.match,
            value="23",
            req='<range-in> { 10 20 }',
        )

    def test_specs_fails_not_opening_brackets_with_op_rangein(self):
        self.assertRaises(
            TypeError,
            specs_matcher.match,
            value="23",
            req='<range-in> ) 10 20 )',
        )

    def test_specs_fails_not_closing_brackets_with_op_rangein(self):
        self.assertRaises(
            TypeError,
            specs_matcher.match,
            value="23",
            req='<range-in> ( 10 20 (',
        )

    def test_specs_fails_reverse_brackets_with_op_rangein(self):
        self.assertRaises(
            TypeError,
            specs_matcher.match,
            value="23",
            req='<range-in> ) 10 20 (',
        )

    def test_specs_fails_too_many_args_with_op_rangein(self):
        self.assertRaises(
            TypeError,
            specs_matcher.match,
            value="23",
            req='<range-in> [ 10 20 30 ]',
        )

    def test_specs_fails_bad_limits_with_op_rangein(self):
        self.assertRaises(
            TypeError,
            specs_matcher.match,
            value="23",
            req='<range-in> [ 20 10 ]',
        )

    def test_specs_fails_match_beyond_scope_with_op_rangein_le(self):
        self._do_specs_matcher_test(
            matches=False, value="23", req='<range-in> [ 10 20 ]'
        )

    def test_specs_fails_match_beyond_scope_with_op_rangein_lt(self):
        self._do_specs_matcher_test(
            matches=False, value="23", req='<range-in> [ 10 20 )'
        )

    def test_specs_fails_match_under_scope_with_op_rangein_ge(self):
        self._do_specs_matcher_test(
            matches=False, value="5", req='<range-in> [ 10 20 ]'
        )

    def test_specs_fails_match_under_scope_with_op_rangein_gt(self):
        self._do_specs_matcher_test(
            matches=False, value="5", req='<range-in> ( 10 20 ]'
        )

    def test_specs_fails_match_float_beyond_scope_with_op_rangein_le(self):
        self._do_specs_matcher_test(
            matches=False, value="20.3", req='<range-in> [ 10.1 20.2 ]'
        )

    def test_specs_fails_match_float_beyond_scope_with_op_rangein_lt(self):
        self._do_specs_matcher_test(
            matches=False, value="20.3", req='<range-in> [ 10.1 20.2 )'
        )

    def test_specs_fails_match_float_under_scope_with_op_rangein_ge(self):
        self._do_specs_matcher_test(
            matches=False, value="5.0", req='<range-in> [ 5.1 20.2 ]'
        )

    def test_specs_fails_match_float_under_scope_with_op_rangein_gt(self):
        self._do_specs_matcher_test(
            matches=False, value="5.0", req='<range-in> ( 5.1 20.2 ]'
        )

    def test_specs_matches_int_lower_int_range_with_op_rangein_ge(self):
        self._do_specs_matcher_test(
            matches=True, value="10", req='<range-in> [ 10 20 ]'
        )

    def test_specs_fails_matchesint_lower_int_range_with_op_rangein_gt(self):
        self._do_specs_matcher_test(
            matches=False, value="10", req='<range-in> ( 10 20 ]'
        )

    def test_specs_matches_float_lower_float_range_with_op_rangein_ge(self):
        self._do_specs_matcher_test(
            matches=True, value="10.1", req='<range-in> [ 10.1 20 ]'
        )

    def test_specs_fails_matche_float_lower_float_range_with_op_rangein_gt(
        self,
    ):
        self._do_specs_matcher_test(
            matches=False, value="10.1", req='<range-in> ( 10.1 20 ]'
        )

    def test_specs_matches_int_with_int_range_with_op_rangein(self):
        self._do_specs_matcher_test(
            matches=True, value="15", req='<range-in> [ 10 20 ]'
        )

    def test_specs_matches_float_with_int_limit_with_op_rangein(self):
        self._do_specs_matcher_test(
            matches=True, value="15.5", req='<range-in> [ 10 20 ]'
        )

    def test_specs_matches_int_upper_int_range_with_op_rangein(self):
        self._do_specs_matcher_test(
            matches=True, value="20", req='<range-in> [ 10 20 ]'
        )

    def test_specs_fails_matche_int_upper_int_range_with_op_rangein_lt(self):
        self._do_specs_matcher_test(
            matches=False, value="20", req='<range-in> [ 10 20 )'
        )

    def test_specs_matches_float_upper_mixed_range_with_op_rangein(self):
        self._do_specs_matcher_test(
            matches=True, value="20.5", req='<range-in> [ 10 20.5 ]'
        )

    def test_specs_fails_matche_float_upper_mixed_range_with_op_rangein_lt(
        self,
    ):
        self._do_specs_matcher_test(
            matches=False, value="20.5", req='<range-in> [ 10 20.5 )'
        )

    def test_specs_matches_float_with_float_limit_with_op_rangein(self):
        self._do_specs_matcher_test(
            matches=True, value="12.5", req='<range-in> [ 10.1 20.1 ]'
        )

    def test_specs_matches_only_one_with_op_rangein(self):
        self._do_specs_matcher_test(
            matches=True, value="10.1", req='<range-in> [ 10.1 10.1 ]'
        )



# Copyright 2012 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Network-related utilities and helper functions.
"""

import ipaddress
import logging
import os
import re
import socket
from typing import cast, Any
from urllib import parse

import netaddr
from netaddr.core import INET_ATON
from netaddr.core import INET_PTON
import psutil

from oslo_utils._i18n import _


LOG = logging.getLogger(__name__)
_IS_IPV6_ENABLED = None


def parse_host_port(
    address: str, default_port: int | None = None
) -> tuple[str | None, int | None]:
    """Interpret a string as a host:port pair.

    An IPv6 address MUST be escaped if accompanied by a port,
    because otherwise ambiguity ensues: 2001:db8:85a3::8a2e:370:7334
    means both [2001:db8:85a3::8a2e:370:7334] and
    [2001:db8:85a3::8a2e:370]:7334.

    >>> parse_host_port('server01:80')
    ('server01', 80)
    >>> parse_host_port('server01')
    ('server01', None)
    >>> parse_host_port('server01', default_port=1234)
    ('server01', 1234)
    >>> parse_host_port('[::1]:80')
    ('::1', 80)
    >>> parse_host_port('[::1]')
    ('::1', None)
    >>> parse_host_port('[::1]', default_port=1234)
    ('::1', 1234)
    >>> parse_host_port('2001:db8:85a3::8a2e:370:7334', default_port=1234)
    ('2001:db8:85a3::8a2e:370:7334', 1234)
    >>> parse_host_port(None)
    (None, None)
    """
    if not address:
        return (None, None)

    port: int | str | None

    if address[0] == '[':
        # Escaped ipv6
        _host, _port = address[1:].split(']')
        host = _host
        if ':' in _port:
            port = int(_port.split(':')[1])
        else:
            port = default_port
    else:
        if address.count(':') == 1:
            host, port = address.split(':')
        else:
            # 0 means ipv4, >1 means ipv6.
            # We prohibit unescaped ipv6 addresses with port.
            host = address
            port = default_port

    return (host, None if port is None else int(port))


def is_valid_ipv4(address: str | None, strict: bool = True) -> bool:
    """Verify that address represents a valid IPv4 address.

    :param address: Value to verify
    :type address: string
    :param strict: flag allowing users to restrict validation
        to IP addresses in presentation format (``a.b.c.d``) as opposed to
        address format (``a.b.c.d``, ``a.b.c``, ``a.b``, ``a``).
    :type flags: bool
    :returns: bool

    .. versionadded:: 1.1
    .. versionchanged:: 4.8.0
       Allow to restrict validation to IP addresses in presentation format
       (``a.b.c.d``) as opposed to address format
       (``a.b.c.d``, ``a.b.c``, ``a.b``, ``a``).
    """
    if not address:
        return False

    flag = INET_PTON if strict else INET_ATON
    try:
        return netaddr.valid_ipv4(address, flags=flag)
    except netaddr.AddrFormatError:
        return False


def is_valid_ipv6(address: str | None) -> bool:
    """Verify that address represents a valid IPv6 address.

    :param address: Value to verify
    :type address: string
    :returns: bool

    .. versionadded:: 1.1
    """
    if not address:
        return False

    parts = address.rsplit("%", 1)
    address = parts[0]
    scope = parts[1] if len(parts) > 1 else None
    if scope is not None and (len(scope) < 1 or len(scope) > 15):
        return False

    try:
        return netaddr.valid_ipv6(address, netaddr.core.INET_PTON)
    except netaddr.AddrFormatError:
        return False


def get_noscope_ipv6(address: str) -> str:
    """Take an IPv6 address and trim scope out if present.

    :param address: Value to change
    :type address: string
    :returns: string

    .. versionadded: 7.3.0:
    """
    _ipv6 = ipaddress.IPv6Address(address)
    if _ipv6.scope_id:
        address = address.removesuffix('%' + _ipv6.scope_id)
    return address


def is_valid_cidr(address: str) -> bool:
    """Verify that address represents a valid CIDR address.

    :param address: Value to verify
    :type address: string
    :returns: bool

    .. versionadded:: 3.8
    """
    try:
        # Validate the correct CIDR Address
        netaddr.IPNetwork(address)
    except (TypeError, netaddr.AddrFormatError):
        return False

    # Prior validation partially verify /xx part
    # Verify it here
    ip_segment = address.split('/')

    if len(ip_segment) <= 1 or ip_segment[1] == '':
        return False

    return True


def is_valid_ipv6_cidr(address: str) -> bool:
    """Verify that address represents a valid IPv6 CIDR address.

    :param address: address to verify
    :type address: string
    :returns: true if address is valid, false otherwise

    .. versionadded:: 3.17
    """
    try:
        netaddr.IPNetwork(address, version=6).cidr
        return True
    except (TypeError, netaddr.AddrFormatError):
        return False


def get_ipv6_addr_by_EUI64(prefix: str, mac: str) -> netaddr.IPAddress:
    """Calculate IPv6 address using EUI-64 specification.

    This method calculates the IPv6 address using the EUI-64
    addressing scheme as explained in rfc2373.

    :param prefix: IPv6 prefix.
    :param mac: IEEE 802 48-bit MAC address.
    :returns: IPv6 address on success.
    :raises ValueError, TypeError: For any invalid input.

    .. versionadded:: 1.4
    """
    if not isinstance(prefix, str):
        msg = _("Prefix must be a string")
        raise TypeError(msg)
    # Check if the prefix is an IPv4 address
    if is_valid_ipv4(prefix, False):
        msg = _("Unable to generate IP address by EUI64 for IPv4 prefix")
        raise ValueError(msg)
    try:
        eui64 = int(netaddr.EUI(mac).eui64())
        return netaddr.IPAddress(
            netaddr.IPNetwork(prefix).first + eui64 ^ (1 << 57)
        )
    except (ValueError, netaddr.AddrFormatError):
        raise ValueError(
            _(
                'Bad prefix or mac format for generating IPv6 '
                'address by EUI-64: %(prefix)s, %(mac)s:'
            )
            % {'prefix': prefix, 'mac': mac}
        )
    except TypeError:
        raise TypeError(
            _('Bad prefix type for generating IPv6 address by EUI-64: %s')
            % prefix
        )


def get_mac_addr_by_ipv6(
    ipv6: netaddr.IPAddress,
    dialect: type[netaddr.mac_eui48] = netaddr.mac_unix_expanded,
) -> netaddr.EUI:
    """Extract MAC address from interface identifier based IPv6 address.

    For example from link-local addresses (fe80::/10) generated from MAC.

    :param ipv6: An interface identifier (i.e. mostly MAC) based IPv6
                 address as a netaddr.IPAddress() object.
    :param dialect: The netaddr dialect of the the object returned.
                    Defaults to netaddr.mac_unix_expanded.
    :returns: A MAC address as a netaddr.EUI() object.

    See also:
    * https://tools.ietf.org/html/rfc4291#appendix-A
    * https://tools.ietf.org/html/rfc4291#section-2.5.6

    .. versionadded:: 4.3.0
    """
    return netaddr.EUI(
        int(
            # out of the lowest 8 bytes (byte positions 8-1)
            # delete the middle 2 bytes (5-4, 0xff_fe)
            # by shifting the highest 3 bytes to the right by 2 bytes (8-6 -> 6-4)
            (
                ((ipv6 & 0xFF_FF_FF_00_00_00_00_00) >> 16)
                +
                # adding the lowest 3 bytes as they are (3-1)
                (ipv6 & 0xFF_FF_FF)  # type: ignore
            )
            ^
            # then invert the universal/local bit
            0x02_00_00_00_00_00
        ),
        dialect=dialect,
    )


def is_ipv6_enabled() -> bool:
    """Check if IPv6 support is enabled on the platform.

    This api will look into the proc entries of the platform to figure
    out the status of IPv6 support on the platform.

    :returns: True if the platform has IPv6 support, False otherwise.

    .. versionadded:: 1.4
    """

    global _IS_IPV6_ENABLED

    if _IS_IPV6_ENABLED is None:
        disabled_ipv6_path = "/proc/sys/net/ipv6/conf/default/disable_ipv6"
        if os.path.exists(disabled_ipv6_path):
            with open(disabled_ipv6_path) as f:
                disabled = f.read().strip()
            _IS_IPV6_ENABLED = disabled == "0"
        else:
            _IS_IPV6_ENABLED = False
    return _IS_IPV6_ENABLED


def escape_ipv6(address: str) -> str:
    """Escape an IP address in square brackets if IPv6

    :param address: address to optionaly escape
    :type address: string
    :returns: string

    .. versionadded:: 3.29.0
    """
    if is_valid_ipv6(address):
        return f"[{address}]"
    return address


def is_valid_ip(address: str | None) -> bool:
    """Verify that address represents a valid IP address.

    :param address: Value to verify
    :type address: string
    :returns: bool

    .. versionadded:: 1.1
    """
    return is_valid_ipv4(address, False) or is_valid_ipv6(address)


def is_valid_mac(address: str) -> bool:
    """Verify the format of a MAC address.

    Check if a MAC address is valid and contains six octets. Accepts
    colon-separated format only.

    :param address: MAC address to be validated.
    :returns: True if valid. False if not.

    .. versionadded:: 3.17
    """
    m = "[0-9a-f]{2}(:[0-9a-f]{2}){5}$"
    return bool(isinstance(address, str) and re.match(m, address.lower()))


def _is_int_in_range(value: int, start: int, end: int) -> bool:
    """Try to convert value to int and check if it lies within
    range 'start' to 'end'.

    :param value: value to verify
    :param start: start number of range
    :param end: end number of range
    :returns: bool
    """
    try:
        val = int(value)
    except (ValueError, TypeError):
        return False
    return start <= val <= end


def is_valid_port(port: int) -> bool:
    """Verify that port represents a valid port number.

    Port can be valid integer having a value of 0 up to and
    including 65535.

    .. versionadded:: 1.1.1
    """
    return _is_int_in_range(port, 0, 65535)


def is_valid_icmp_type(type: int) -> bool:
    """Verify if ICMP type is valid.

    :param type: ICMP *type* field can only be a valid integer
    :returns: bool

    ICMP *type* field can be valid integer having a value of 0
    up to and including 255.
    """
    return _is_int_in_range(type, 0, 255)


def is_valid_icmp_code(code: int | None) -> bool:
    """Verify if ICMP code is valid.

    :param code: ICMP *code* field can be valid integer or None
    :returns: bool

    ICMP *code* field can be either None or valid integer having
    a value of 0 up to and including 255.
    """
    if code is None:
        return True
    return _is_int_in_range(code, 0, 255)


def get_my_ipv4() -> str:
    """Returns the actual ipv4 of the local machine.

    This code figures out what source address would be used if some traffic
    were to be sent out to some well known address on the Internet. In this
    case, IP from RFC5737 is used, but the specific address does not
    matter much. No traffic is actually sent.

    .. versionadded:: 1.1

    .. versionchanged:: 1.2.1
       Return ``'127.0.0.1'`` if there is no default interface.
    """
    try:
        with socket.socket(socket.AF_INET, socket.SOCK_DGRAM) as csock:
            csock.connect(('192.0.2.0', 80))
            return cast(str, (csock.getsockname()[0]))
    except OSError:
        return _get_my_ipv4_address()


def _get_my_ipv4_address() -> str:
    """Figure out the best ipv4"""
    LOCALHOST = '127.0.0.1'
    interface = None

    try:
        with open('/proc/net/route') as routes:
            for route in routes:
                route_attrs = route.strip().split('\t')
                if route_attrs[1] == '00000000':
                    interface = route_attrs[0]
                    break
            else:
                LOG.info(
                    'Could not determine default network interface, '
                    'using %s for IPv4 address',
                    LOCALHOST,
                )
                return LOCALHOST
    except FileNotFoundError:
        LOG.info(
            'IPv4 route table not found, using %s for IPv4 address', LOCALHOST
        )
        return LOCALHOST

    try:
        addrs = psutil.net_if_addrs()[interface]
        v4addrs = [addr for addr in addrs if addr.family == socket.AF_INET]
        return v4addrs[0].address
    except (KeyError, IndexError):
        LOG.info(
            'Could not determine IPv4 address for interface %s, '
            'using 127.0.0.1',
            interface,
        )
    except Exception as e:
        LOG.info(
            'Could not determine IPv4 address for '
            'interface %(interface)s: %(error)s',
            {'interface': interface, 'error': e},
        )
    return LOCALHOST


def get_my_ipv6() -> str:
    """Returns the actual IPv6 address of the local machine.

    This code figures out what source address would be used if some traffic
    were to be sent out to some well known address on the Internet. In this
    case, IPv6 from RFC3849 is used, but the specific address does not
    matter much. No traffic is actually sent.

    .. versionadded:: 6.1
       Return ``'::1'`` if there is no default interface.
    """
    try:
        with socket.socket(socket.AF_INET6, socket.SOCK_DGRAM) as csock:
            csock.connect(('2001:db8::1', 80))
            return cast(str, csock.getsockname()[0])
    except OSError:
        return _get_my_ipv6_address()


def _get_my_ipv6_address() -> str:
    """Figure out the best IPv6 address"""
    LOCALHOST = '::1'
    interface = None
    ZERO_ADDRESS = '00000000000000000000000000000000'
    ZERO_PAIR = (ZERO_ADDRESS, '00')

    try:
        with open('/proc/net/ipv6_route') as routes:
            for route in routes:
                route_attrs = route.strip().split(' ')
                if (route_attrs[0], route_attrs[1]) == ZERO_PAIR and (
                    route_attrs[2],
                    route_attrs[3],
                ) == ZERO_PAIR:
                    interface = route_attrs[-1]
                    break
            else:
                LOG.info(
                    'Could not determine default network interface, '
                    'using %s for IPv6 address',
                    LOCALHOST,
                )
                return LOCALHOST
    except FileNotFoundError:
        LOG.info(
            'IPv6 route table not found, using %s for IPv6 address', LOCALHOST
        )
        return LOCALHOST

    try:
        addrs = psutil.net_if_addrs()[interface]
        v6addrs = [addr for addr in addrs if addr.family == socket.AF_INET6]
        return v6addrs[0].address
    except (KeyError, IndexError):
        LOG.info(
            'Could not determine IPv6 address for interface '
            '%(interface)s, using %(address)s',
            {'interface': interface, 'address': LOCALHOST},
        )
    except Exception as e:
        LOG.info(
            'Could not determine IPv6 address for '
            'interface %(interface)s: %(error)s',
            {'interface': interface, 'error': e},
        )
    return LOCALHOST


class _ModifiedSplitResult(parse.SplitResult):
    """Split results class for urlsplit."""

    def params(self, collapse: bool = True) -> dict[str, list[str] | str]:
        """Extracts the query parameters from the split urls components.

        This method will provide back as a dictionary the query parameter
        names and values that were provided in the url.

        :param collapse: Boolean, turn on or off collapsing of query values
        with the same name. Since a url can contain the same query parameter
        name with different values it may or may not be useful for users to
        care that this has happened. This parameter when True uses the
        last value that was given for a given name, while if False it will
        retain all values provided by associating the query parameter name with
        a list of values instead of a single (non-list) value.
        """
        if self.query:
            if collapse:
                return dict(parse.parse_qsl(self.query))
            else:
                params: dict[str, list[str] | str] = {}
                for key, value in parse.parse_qsl(self.query):
                    if key in params:
                        v = params[key]
                        if isinstance(v, list):
                            v.append(value)
                        else:
                            params[key] = [v, value]
                    else:
                        params[key] = value
                return params
        else:
            return {}


def urlsplit(url: str, scheme: str = '', allow_fragments: bool = True) -> Any:
    """Parse a URL using urlparse.urlsplit(), splitting query and fragments.
    This function papers over Python issue9374_ when needed.

    .. _issue9374: http://bugs.python.org/issue9374

    The parameters are the same as urlparse.urlsplit.
    """
    scheme, netloc, path, query, fragment = parse.urlsplit(
        url, scheme, allow_fragments
    )
    if allow_fragments and '#' in path:
        path, fragment = path.split('#', 1)
    if '?' in path:
        path, query = path.split('?', 1)
    return _ModifiedSplitResult(scheme, netloc, path, query, fragment)


def set_tcp_keepalive(
    sock: socket.socket,
    tcp_keepalive: bool = True,
    tcp_keepidle: int | None = None,
    tcp_keepalive_interval: int | None = None,
    tcp_keepalive_count: int | None = None,
) -> None:
    """Set values for tcp keepalive parameters

    This function configures tcp keepalive parameters if users wish to do
    so.

    :param tcp_keepalive: Boolean, turn on or off tcp_keepalive. If users are
      not sure, this should be True, and default values will be used.

    :param tcp_keepidle: time to wait before starting to send keepalive probes
    :param tcp_keepalive_interval: time between successive probes, once the
      initial wait time is over
    :param tcp_keepalive_count: number of probes to send before the connection
      is killed
    """

    # NOTE(praneshp): Despite keepalive being a tcp concept, the level is
    # still SOL_SOCKET. This is a quirk.
    if isinstance(tcp_keepalive, bool):
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, tcp_keepalive)
    else:
        raise TypeError("tcp_keepalive must be a boolean")

    if not tcp_keepalive:
        return

    # These options aren't available in the OS X version of eventlet,
    # Idle + Count * Interval effectively gives you the total timeout.
    if tcp_keepidle is not None:
        if hasattr(socket, 'TCP_KEEPIDLE'):
            sock.setsockopt(
                socket.IPPROTO_TCP, socket.TCP_KEEPIDLE, tcp_keepidle
            )
        else:
            LOG.warning('tcp_keepidle not available on your system')
    if tcp_keepalive_interval is not None:
        if hasattr(socket, 'TCP_KEEPINTVL'):
            sock.setsockopt(
                socket.IPPROTO_TCP,
                socket.TCP_KEEPINTVL,
                tcp_keepalive_interval,
            )
        else:
            LOG.warning('tcp_keepintvl not available on your system')
    if tcp_keepalive_count is not None:
        if hasattr(socket, 'TCP_KEEPCNT'):
            sock.setsockopt(
                socket.IPPROTO_TCP, socket.TCP_KEEPCNT, tcp_keepalive_count
            )
        else:
            LOG.warning('tcp_keepcnt not available on your system')



# Copyright (c) 2013 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslotest import base as test_base

from oslo_utils import versionutils


class IsCompatibleTestCase(test_base.BaseTestCase):
    def test_same_version(self):
        self.assertTrue(versionutils.is_compatible('1', '1'))
        self.assertTrue(versionutils.is_compatible('1.0', '1.0'))
        self.assertTrue(versionutils.is_compatible('1.0.0', '1.0.0'))

    def test_requested_minor_greater(self):
        self.assertFalse(versionutils.is_compatible('1.1', '1.0'))

    def test_requested_minor_less_than(self):
        self.assertTrue(versionutils.is_compatible('1.0', '1.1'))

    def test_requested_patch_greater(self):
        self.assertFalse(versionutils.is_compatible('1.0.1', '1.0.0'))

    def test_requested_patch_less_than(self):
        self.assertTrue(versionutils.is_compatible('1.0.0', '1.0.1'))

    def test_requested_patch_not_present_same(self):
        self.assertTrue(versionutils.is_compatible('1.0', '1.0.0'))

    def test_requested_patch_not_present_less_than(self):
        self.assertTrue(versionutils.is_compatible('1.0', '1.0.1'))

    def test_current_patch_not_present_same(self):
        self.assertTrue(versionutils.is_compatible('1.0.0', '1.0'))

    def test_current_patch_not_present_less_than(self):
        self.assertFalse(versionutils.is_compatible('1.0.1', '1.0'))

    def test_same_major_true(self):
        """Even though the current version is 2.0, since `same_major` defaults
        to `True`, 1.0 is deemed incompatible.
        """
        self.assertFalse(versionutils.is_compatible('2.0', '1.0'))
        self.assertTrue(versionutils.is_compatible('1.0', '1.0'))
        self.assertFalse(versionutils.is_compatible('1.0', '2.0'))

    def test_same_major_false(self):
        """With `same_major` set to False, then major version compatibiity
        rule is not enforced, so a current version of 2.0 is deemed to satisfy
        a requirement of 1.0.
        """
        self.assertFalse(
            versionutils.is_compatible('2.0', '1.0', same_major=False)
        )
        self.assertTrue(
            versionutils.is_compatible('1.0', '1.0', same_major=False)
        )
        self.assertTrue(
            versionutils.is_compatible('1.0', '2.0', same_major=False)
        )

    def test_convert_version_to_int(self):
        self.assertEqual(6002000, versionutils.convert_version_to_int('6.2.0'))
        self.assertEqual(
            6004003, versionutils.convert_version_to_int((6, 4, 3))
        )
        self.assertEqual(5, versionutils.convert_version_to_int((5,)))
        self.assertRaises(
            ValueError, versionutils.convert_version_to_int, '5a.6b'
        )

    def test_convert_version_to_string(self):
        self.assertEqual('6.7.0', versionutils.convert_version_to_str(6007000))
        self.assertEqual('4', versionutils.convert_version_to_str(4))

    def test_convert_version_to_tuple(self):
        self.assertEqual(
            (6, 7, 0), versionutils.convert_version_to_tuple('6.7.0')
        )
        self.assertEqual(
            (6, 7, 0), versionutils.convert_version_to_tuple('6.7.0a1')
        )
        self.assertEqual(
            (6, 7, 0), versionutils.convert_version_to_tuple('6.7.0alpha1')
        )
        self.assertEqual(
            (6, 7, 0), versionutils.convert_version_to_tuple('6.7.0b1')
        )
        self.assertEqual(
            (6, 7, 0), versionutils.convert_version_to_tuple('6.7.0beta1')
        )
        self.assertEqual(
            (6, 7, 0), versionutils.convert_version_to_tuple('6.7.0rc1')
        )


class VersionPredicateTest(test_base.BaseTestCase):
    def test_version_predicate_valid(self):
        pred = versionutils.VersionPredicate('<2.0.0')
        self.assertTrue(pred.satisfied_by('1.10.0'))
        self.assertFalse(pred.satisfied_by('2.0.0'))

        pred = versionutils.VersionPredicate('<=2.0.0')
        self.assertTrue(pred.satisfied_by('2.0.0'))
        self.assertFalse(pred.satisfied_by('2.1.0'))

        pred = versionutils.VersionPredicate('>2.0.0')
        self.assertFalse(pred.satisfied_by('1.10.0'))
        self.assertTrue(pred.satisfied_by('2.1.0'))

        pred = versionutils.VersionPredicate('>=2.0.0')
        self.assertFalse(pred.satisfied_by('1.9.0'))
        self.assertTrue(pred.satisfied_by('2.0.0'))

        pred = versionutils.VersionPredicate('== 2.0.0')
        self.assertFalse(pred.satisfied_by('1.9.9'))
        self.assertTrue(pred.satisfied_by('2.0.0'))
        self.assertFalse(pred.satisfied_by('2.0.1'))

        pred = versionutils.VersionPredicate('!= 2.0.0')
        self.assertTrue(pred.satisfied_by('1.9.9'))
        self.assertFalse(pred.satisfied_by('2.0.0'))
        self.assertTrue(pred.satisfied_by('2.0.1'))

    def test_version_predicate_valid_multi(self):
        pred = versionutils.VersionPredicate('<3.0.0,>=2.1.0')
        self.assertTrue(pred.satisfied_by('2.1.0'))
        self.assertTrue(pred.satisfied_by('2.11.0'))
        self.assertFalse(pred.satisfied_by('2.0.0'))
        self.assertFalse(pred.satisfied_by('3.0.0'))

        pred = versionutils.VersionPredicate(' < 2.0.0, >= 1.0.0 ')
        self.assertTrue(pred.satisfied_by('1.0.0'))
        self.assertTrue(pred.satisfied_by('1.10.0'))
        self.assertFalse(pred.satisfied_by('0.9.0'))
        self.assertFalse(pred.satisfied_by('2.0.0'))

    def test_version_predicate_valid_invalid(self):
        for invalid_str in [
            '3.0.0',
            'foo',
            '<> 3.0.0',
            '>=1.0.0;<2.0.0',
            '>abc',
            '>=1.0.0,',
            '>=1.0.0,2.0.0',
        ]:
            self.assertRaises(
                ValueError, versionutils.VersionPredicate, invalid_str
            )



# Copyright 2011 OpenStack Foundation.
# Copyright 2012, Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Exception related utilities.
"""

from collections.abc import Callable
import functools
import io
import logging
import os
import sys
import time
import traceback
import types
from typing import Any

from oslo_utils import encodeutils
from oslo_utils import reflection
from oslo_utils import timeutils


class CausedByException(Exception):
    """Base class for exceptions which have associated causes.

    NOTE(harlowja): in later versions of python we can likely remove the need
    to have a ``cause`` here as PY3+ have implemented :pep:`3134` which
    handles chaining in a much more elegant manner.

    :param message: the exception message, typically some string that is
                    useful for consumers to view when debugging or analyzing
                    failures.
    :param cause: the cause of the exception being raised, when provided this
                  should itself be an exception instance, this is useful for
                  creating a chain of exceptions for versions of python where
                  this is not yet implemented/supported natively.

    .. versionadded:: 2.4
    """

    def __init__(self, message: str, cause: Exception | None = None) -> None:
        super().__init__(message)
        self.cause = cause

    def __bytes__(self) -> bytes:
        return self.pformat().encode("utf8")

    def __str__(self) -> str:
        return self.pformat()

    def _get_message(self) -> Any:
        # We must *not* call into the ``__str__`` method as that will
        # reactivate the pformat method, which will end up badly (and doesn't
        # look pretty at all); so be careful...
        return self.args[0]

    def pformat(
        self,
        indent: int = 2,
        indent_text: str = " ",
        show_root_class: bool = False,
    ) -> str:
        """Pretty formats a caused exception + any connected causes."""
        if indent < 0:
            raise ValueError(
                "Provided 'indent' must be greater than"
                f" or equal to zero instead of {indent}"
            )
        buf = io.StringIO()
        if show_root_class:
            buf.write(reflection.get_class_name(self, fully_qualified=False))
            buf.write(": ")
        buf.write(self._get_message())
        active_indent = indent
        next_up = self.cause
        seen = []
        while next_up is not None and next_up not in seen:
            seen.append(next_up)
            buf.write(os.linesep)
            if isinstance(next_up, CausedByException):
                buf.write(indent_text * active_indent)
                buf.write(
                    reflection.get_class_name(next_up, fully_qualified=False)
                )
                buf.write(": ")
                buf.write(next_up._get_message())
            else:
                lines = traceback.format_exception_only(type(next_up), next_up)
                for i, line in enumerate(lines):
                    buf.write(indent_text * active_indent)
                    if line.endswith("\n"):
                        # We'll add our own newlines on...
                        line = line[0:-1]
                    buf.write(line)
                    if i + 1 != len(lines):
                        buf.write(os.linesep)
                # Don't go deeper into non-caused-by exceptions... as we
                # don't know if there exception 'cause' attributes are even
                # useable objects...
                break
            active_indent += indent
            next_up = getattr(next_up, 'cause', None)
        return buf.getvalue()


def raise_with_cause(
    exc_cls: type[Exception], message: str, *args: Any, **kwargs: Any
) -> None:
    """Helper to raise + chain exceptions (when able) and associate a *cause*.

    NOTE(harlowja): Since in py3.x exceptions can be chained (due to
    :pep:`3134`) we should try to raise the desired exception with the given
    *cause* (or extract a *cause* from the current stack if able) so that the
    exception formats nicely in old and new versions of python. Since py2.x
    does **not** support exception chaining (or formatting) the exception
    class provided should take a ``cause`` keyword argument (which it may
    discard if it wants) to its constructor which can then be
    inspected/retained on py2.x to get *similar* information as would be
    automatically included/obtainable in py3.x.

    :param exc_cls: the exception class to raise (typically one derived
                    from :py:class:`.CausedByException` or equivalent).
    :param message: the text/str message that will be passed to
                    the exceptions constructor as its first positional
                    argument.
    :param args: any additional positional arguments to pass to the
                 exceptions constructor.
    :param kwargs: any additional keyword arguments to pass to the
                   exceptions constructor.

    .. versionadded:: 1.6
    """
    if 'cause' not in kwargs:
        exc_type, exc, exc_tb = sys.exc_info()
        try:
            if exc is not None:
                kwargs['cause'] = exc
        finally:
            # Leave no references around (especially with regards to
            # tracebacks and any variables that it retains internally).
            del (exc_type, exc, exc_tb)
    raise exc_cls(message, *args, **kwargs) from kwargs.get('cause')


class save_and_reraise_exception:
    """Save current exception, run some code and then re-raise.

    In some cases the exception context can be cleared, resulting in None
    being attempted to be re-raised after an exception handler is run. This
    can happen when eventlet switches greenthreads or when running an
    exception handler, code raises and catches an exception. In both
    cases the exception context will be cleared.

    To work around this, we save the exception state, run handler code, and
    then re-raise the original exception. If another exception occurs, the
    saved exception is logged and the new exception is re-raised.

    In some cases the caller may not want to re-raise the exception, and
    for those circumstances this context provides a reraise flag that
    can be used to suppress the exception.  For example::

      except Exception:
          with save_and_reraise_exception() as ctxt:
              decide_if_need_reraise()
              if not should_be_reraised:
                  ctxt.reraise = False

    If another exception occurs and reraise flag is False,
    the saved exception will not be logged.

    If the caller wants to raise new exception during exception handling
    he/she sets reraise to False initially with an ability to set it back to
    True if needed::

      except Exception:
          with save_and_reraise_exception(reraise=False) as ctxt:
              [if statements to determine whether to raise a new exception]
              # Not raising a new exception, so reraise
              ctxt.reraise = True

    .. versionchanged:: 1.4
       Added *logger* optional parameter.
    """

    def __init__(
        self, reraise: bool = True, logger: logging.Logger | None = None
    ) -> None:
        self.reraise = reraise
        if logger is None:
            logger = logging.getLogger()
        self.logger = logger
        self.type_: type[BaseException] | None = None
        self.value: BaseException | None = None
        self.tb: types.TracebackType | None = None

    def force_reraise(self) -> None:
        if self.type_ is None and self.value is None:
            raise RuntimeError(
                "There is no (currently) captured exception"
                " to force the reraising of"
            )
        try:
            if self.value is None and self.type_ is not None:
                self.value = self.type_()

            if self.value is not None:
                if self.value.__traceback__ is not self.tb:
                    raise self.value.with_traceback(self.tb)
                raise self.value
        finally:
            self.value = None
            self.tb = None

    def capture(self, check: bool = True) -> 'save_and_reraise_exception':
        type_, value, tb = sys.exc_info()
        if check and type_ is None and value is None:
            raise RuntimeError("There is no active exception to capture")
        self.type_, self.value, self.tb = (type_, value, tb)
        return self

    def __enter__(self) -> 'save_and_reraise_exception':
        # TODO(harlowja): perhaps someday in the future turn check here
        # to true, because that is likely the desired intention, and doing
        # so ensures that people are actually using this correctly.
        return self.capture(check=False)

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: types.TracebackType | None,
    ) -> bool | None:
        if exc_type is not None:
            if self.reraise:
                self.logger.error(
                    'Original exception being dropped: %s',
                    traceback.format_exception(
                        self.type_, self.value, self.tb
                    ),
                )
            return False
        if self.reraise:
            self.force_reraise()
        return None


def forever_retry_uncaught_exceptions(
    *args: Any, **kwargs: Any
) -> Callable[[Callable[..., Any]], Callable[..., Any]]:
    """Decorates provided function with infinite retry behavior.

    The function retry delay is **always** one second unless
    keyword argument ``retry_delay`` is passed that defines a value different
    than 1.0 (less than zero values are automatically changed to be 0.0).

    If repeated exceptions with the same message occur, logging will only
    output/get triggered for those equivalent messages every 60.0
    seconds, this can be altered by keyword argument ``same_log_delay`` to
    be a value different than 60.0 seconds (exceptions that change the
    message are always logged no matter what this delay is set to). As in
    the ``retry_delay`` case if this is less than zero, it will be
    automatically changed to be 0.0.
    """

    def decorator(infunc: Callable[..., Any]) -> Callable[..., Any]:
        retry_delay = max(0.0, float(kwargs.get('retry_delay', 1.0)))
        same_log_delay = max(0.0, float(kwargs.get('same_log_delay', 60.0)))

        @functools.wraps(infunc)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            last_exc_message = None
            same_failure_count = 0
            watch = timeutils.StopWatch(duration=same_log_delay)
            while True:
                try:
                    return infunc(*args, **kwargs)
                except Exception as exc:
                    this_exc_message = encodeutils.exception_to_unicode(exc)
                    if this_exc_message == last_exc_message:
                        same_failure_count += 1
                    else:
                        same_failure_count = 1
                    if this_exc_message != last_exc_message or watch.expired():
                        # The watch has expired or the exception message
                        # changed, so time to log it again...
                        logging.exception(
                            f'Unexpected exception occurred '
                            f'{same_failure_count} time(s)... retrying.'
                        )
                        if not watch.has_started():
                            watch.start()
                        else:
                            watch.restart()
                        same_failure_count = 0
                        last_exc_message = this_exc_message
                    time.sleep(retry_delay)

        return wrapper

    # This is needed to handle when the decorator has args or the decorator
    # doesn't have args, python is rather weird here...
    if kwargs or not args:
        return decorator
    else:
        if len(args) == 1:
            return decorator(args[0])
        else:
            return decorator


class exception_filter:
    """A context manager that prevents some exceptions from being raised.

    Use this class as a decorator for a function that returns whether a given
    exception should be ignored, in cases where complex logic beyond subclass
    matching is required. e.g.

    >>> @exception_filter
    >>> def ignore_test_assertions(ex):
    ...     return isinstance(ex, AssertionError) and 'test' in str(ex)

    The filter matching function can then be used as a context manager:

    >>> with ignore_test_assertions:
    ...     assert False, 'This is a test'

    or called directly:

    >>> try:
    ...     assert False, 'This is a test'
    ... except Exception as ex:
    ...     ignore_test_assertions(ex)

    Any non-matching exception will be re-raised. When the filter is used as a
    context manager, the traceback for re-raised exceptions is always
    preserved. When the filter is called as a function, the traceback is
    preserved provided that no other exceptions have been raised in the
    intervening time. The context manager method is preferred for this reason
    except in cases where the ignored exception affects control flow.
    """

    def __init__(
        self, should_ignore_ex: Callable[[BaseException], bool]
    ) -> None:
        self._should_ignore_ex = should_ignore_ex

        if all(
            hasattr(should_ignore_ex, a) for a in functools.WRAPPER_ASSIGNMENTS
        ):
            functools.update_wrapper(self, should_ignore_ex)

    def __get__(self, obj: Any, owner: Any) -> 'exception_filter':
        return type(self)(self._should_ignore_ex.__get__(obj, owner))

    def __enter__(self) -> 'exception_filter':
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: types.TracebackType | None,
    ) -> bool | None:
        if exc_val is not None:
            return self._should_ignore_ex(exc_val)
        return False

    def __call__(self, ex: BaseException) -> None:
        """Re-raise any exception value not being filtered out.

        If the exception was the last to be raised, it will be re-raised with
        its original traceback.
        """
        exc_type, exc_val, traceback = sys.exc_info()

        try:
            if not self._should_ignore_ex(ex):
                if exc_val is ex:
                    try:
                        if exc_val is None:
                            exc_val = exc_type()
                        if exc_val.__traceback__ is not traceback:
                            raise exc_val.with_traceback(traceback)
                        raise exc_val
                    finally:
                        exc_val = None
                        traceback = None
                else:
                    raise ex
        finally:
            del exc_type, exc_val, traceback



# Copyright 2011 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
System-level utilities and helper functions.
"""

import collections.abc
import math
import re
from typing import Any
import unicodedata
import urllib.parse

from oslo_utils._i18n import _
from oslo_utils import encodeutils


UNIT_PREFIX_EXPONENT = {
    'k': 1,
    'K': 1,
    'Ki': 1,
    'M': 2,
    'Mi': 2,
    'G': 3,
    'Gi': 3,
    'T': 4,
    'Ti': 4,
    'P': 5,
    'Pi': 5,
    'E': 6,
    'Ei': 6,
    'Z': 7,
    'Zi': 7,
    'Y': 8,
    'Yi': 8,
    'R': 9,
    'Ri': 9,
    'Q': 10,
    'Qi': 10,
}
UNIT_SYSTEM_INFO = {
    'IEC': (1024, re.compile(r'(^[-+]?\d*\.?\d+)([KMGTPEZYRQ]i?)?(b|bit|B)$')),
    'SI': (1000, re.compile(r'(^[-+]?\d*\.?\d+)([kMGTPEZYRQ])?(b|bit|B)$')),
    'mixed': (
        None,
        re.compile(r'(^[-+]?\d*\.?\d+)([kKMGTPEZYRQ]i?)?(b|bit|B)$'),
    ),
}

TRUE_STRINGS = ('1', 't', 'true', 'on', 'y', 'yes')
FALSE_STRINGS = ('0', 'f', 'false', 'off', 'n', 'no')

SLUGIFY_STRIP_RE = re.compile(r"[^\w\s-]")
SLUGIFY_HYPHENATE_RE = re.compile(r"[-\s]+")


# NOTE(flaper87): The following globals are used by `mask_password` and
#                 `mask_dict_password`. They must all be lowercase.
_SANITIZE_KEYS = [
    'adminpass',
    'admin_pass',
    'password',
    'admin_password',
    'auth_token',
    'new_pass',
    'auth_password',
    'secret_uuid',
    'secret',
    'sys_pswd',
    'token',
    'configdrive',
    'chappassword',
    'encrypted_key',
    'private_key',
    'fernetkey',
    'sslkey',
    'passphrase',
    'cephclusterfsid',
    'octaviaheartbeatkey',
    'rabbitcookie',
    'cephmanilaclientkey',
    'pacemakerremoteauthkey',
    'designaterndckey',
    'cephadminkey',
    'heatauthencryptionkey',
    'cephclientkey',
    'keystonecredential',
    'barbicansimplecryptokek',
    'cephrgwkey',
    'swifthashsuffix',
    'migrationsshkey',
    'cephmdskey',
    'cephmonkey',
    'chapsecret',
]

# NOTE(ldbragst): Let's build a list of regex objects using the list of
# _SANITIZE_KEYS we already have. This way, we only have to add the new key
# to the list of _SANITIZE_KEYS and we can generate regular expressions
# for XML and JSON automatically.
_SANITIZE_PATTERNS_2: dict[str, Any] = {}
_SANITIZE_PATTERNS_1: dict[str, Any] = {}
_SANITIZE_PATTERNS_WILDCARD: dict[str, Any] = {}

# NOTE(amrith): Some regular expressions have only one parameter, some
# have two parameters. Use different lists of patterns here.
_FORMAT_PATTERNS_1 = [r'(%(key)s[0-9]*\s*[=]\s*)[^\s^\'^\"]+']
_FORMAT_PATTERNS_2 = [
    r'(%(key)s[0-9]*\s*[=]\s*[\"\'])[^\"\']*([\"\'])',
    r'(%(key)s[0-9]*\s*[=]\s*[\"])[^\"]*([\"])',
    r'(%(key)s[0-9]*\s*[=]\s*[\'])[^\']*([\'])',
    r'(%(key)s[0-9]*\s+[\"\'])[^\"\']*([\"\'])',
    r'([-]{2}%(key)s[0-9]*\s+)[^\'^\"^=^\s]+([\s]*)',
    r'(<%(key)s[0-9]*>)[^<]*(</%(key)s[0-9]*>)',
    r'([\"\']%(key)s[0-9]*[\"\']\s*:\s*[\"\'])[^\"\']*'
    r'([\"\'])',
    r'([\'"][^"\']*%(key)s[0-9]*[\'"]\s*:\s*u?[\'"])[^\"\']*'
    r'([\'"])',
    r'([\'"][^\'"]*%(key)s[0-9]*[\'"]\s*,\s*\'--?[A-z]+'
    r'\'\s*,\s*u?[\'"])[^\"\']*([\'"])',
    r'(%(key)s[0-9]*\s*--?[A-z]+\s*)\S+(\s*)',
]
_FORMAT_PATTERNS_WILDCARD = [
    r'([\'\"][^\"\']*%(key)s[0-9]*[\'\"]\s*:\s*u?[\'\"].*[\'\"])[^\"\']*([\'\"])'
]  # noqa: E501

# NOTE(dhellmann): Keep a separate list of patterns by key so we only
# need to apply the substitutions for keys we find using a quick "in"
# test.
for key in _SANITIZE_KEYS:
    _SANITIZE_PATTERNS_1[key] = []
    _SANITIZE_PATTERNS_2[key] = []
    _SANITIZE_PATTERNS_WILDCARD[key] = []

    for pattern in _FORMAT_PATTERNS_2:
        reg_ex = re.compile(pattern % {'key': key}, re.DOTALL | re.IGNORECASE)
        _SANITIZE_PATTERNS_2[key].append(reg_ex)

    for pattern in _FORMAT_PATTERNS_1:
        reg_ex = re.compile(pattern % {'key': key}, re.DOTALL | re.IGNORECASE)
        _SANITIZE_PATTERNS_1[key].append(reg_ex)

    for pattern in _FORMAT_PATTERNS_WILDCARD:
        reg_ex = re.compile(pattern % {'key': key}, re.DOTALL | re.IGNORECASE)
        _SANITIZE_PATTERNS_WILDCARD[key].append(reg_ex)


def int_from_bool_as_string(subject: Any) -> int:
    """Interpret a string as a boolean and return either 1 or 0.

    Any string value in:

        ('True', 'true', 'On', 'on', '1')

    is interpreted as a boolean True.

    Useful for JSON-decoded stuff and config file parsing
    """
    return int(bool_from_string(subject))


def bool_from_string(
    subject: Any, strict: bool = False, default: bool = False
) -> bool:
    """Interpret a subject as a boolean.

    A subject can be a boolean, a string or an integer. Boolean type value
    will be returned directly, otherwise the subject will be converted to
    a string. A case-insensitive match is performed such that strings
    matching 't','true', 'on', 'y', 'yes', or '1' are considered True and,
    when `strict=False`, anything else returns the value specified by
    'default'.

    Useful for JSON-decoded stuff and config file parsing.

    If `strict=True`, unrecognized values, including None, will raise a
    ValueError which is useful when parsing values passed in from an API call.
    Strings yielding False are 'f', 'false', 'off', 'n', 'no', or '0'.
    """
    if isinstance(subject, bool):
        return subject
    if not isinstance(subject, str):
        subject = str(subject)

    lowered = subject.strip().lower()

    if lowered in TRUE_STRINGS:
        return True
    elif lowered in FALSE_STRINGS:
        return False
    elif strict:
        acceptable = ', '.join(
            f"'{s}'" for s in sorted(TRUE_STRINGS + FALSE_STRINGS)
        )
        msg = _(
            "Unrecognized value '%(val)s', acceptable values are:"
            " %(acceptable)s"
        ) % {'val': subject, 'acceptable': acceptable}
        raise ValueError(msg)
    else:
        return default


def is_valid_boolstr(value: Any) -> bool:
    """Check if the provided string is a valid bool string or not.

    :param value: value to verify
    :type value: string
    :returns: true if value is boolean string, false otherwise

    .. versionadded:: 3.17
    """
    boolstrs = TRUE_STRINGS + FALSE_STRINGS
    return str(value).lower() in boolstrs


def string_to_bytes(
    text: str, unit_system: str = 'IEC', return_int: bool = False
) -> int | float:
    """Converts a string into an float representation of bytes.

    The units supported for IEC / mixed::

        (
            Kb(it),
            Kib(it),
            Mb(it),
            Mib(it),
            Gb(it),
            Gib(it),
            Tb(it),
            Tib(it),
        )
        (
            Pb(it),
            Pib(it),
            Eb(it),
            Eib(it),
            Zb(it),
            Zib(it),
            Yb(it),
            Yib(it),
        )
        Rb(it), Rib(it), Qb(it), Qib(it)

        (
            KB,
            KiB,
            MB,
            MiB,
            GB,
            GiB,
            TB,
            TiB,
            PB,
            PiB,
            EB,
            EiB,
            ZB,
            ZiB,
        )
        YB, YiB, RB, RiB, QB, QiB

    The units supported for SI ::

        (
            kb(it),
            Mb(it),
            Gb(it),
            Tb(it),
            Pb(it),
            Eb(it),
            Zb(it),
            Yb(it),
        )
        Rb(it), Qb(it)

        kB, MB, GB, TB, PB, EB, ZB, YB, RB, QB

    SI units are interpreted as power-of-ten (e.g. 1kb = 1000b).  Note
    that the SI unit system does not support capital letter 'K'

    IEC units are interpreted as power-of-two (e.g. 1MiB = 1MB =
    1024b)

    Mixed units interpret the "i" to mean IEC, and no "i" to mean SI
    (e.g. 1kb = 1000b, 1kib == 1024b).  Additionaly, mixed units
    interpret 'K' as power-of-ten.  This mode is not particuarly
    useful for new code, but can help with compatability for parsers
    such as GNU parted.

    :param text: String input for bytes size conversion.
    :param unit_system: Unit system for byte size conversion.
    :param return_int: If True, returns integer representation of text
                       in bytes. (default: decimal)
    :returns: Numerical representation of text in bytes.
    :raises ValueError: If text has an invalid value.

    """
    try:
        base, reg_ex = UNIT_SYSTEM_INFO[unit_system]
    except KeyError:
        msg = _('Invalid unit system: "%s"') % unit_system
        raise ValueError(msg)
    match = reg_ex.match(text)
    if match:
        magnitude = float(match.group(1))
        unit_prefix = match.group(2)
        if match.group(3) in ['b', 'bit']:
            magnitude /= 8

        # In the mixed matcher, IEC units (with a trailing 'i') are
        # interpreted as power-of-two, others as power-of-ten
        if unit_system == 'mixed':
            if unit_prefix and not unit_prefix.endswith('i'):
                # For maximum compatability in mixed mode, we understand
                # "K" (which is not strict SI) as "k"
                if unit_prefix.startswith == 'K':
                    unit_prefix = 'k'
                base = 1000
            else:
                base = 1024
    else:
        msg = _('Invalid string format: %s') % text
        raise ValueError(msg)

    if base is None:
        msg = _('Could not auto-detect base')
        raise ValueError(msg)

    if not unit_prefix:
        res = magnitude
    else:
        exponent = UNIT_PREFIX_EXPONENT.get(unit_prefix)
        if exponent is None:
            msg = _('Invalid unit prefix: %s') % unit_prefix
            raise ValueError(msg)
        res = magnitude * pow(base, exponent)
    if return_int:
        return int(math.ceil(res))
    return res


def to_slug(
    value: Any, incoming: str | None = None, errors: str = "strict"
) -> str:
    """Normalize string.

    Convert to lowercase, remove non-word characters, and convert spaces
    to hyphens.

    Inspired by Django's `slugify` filter.

    :param value: Text to slugify
    :param incoming: Text's current encoding
    :param errors: Errors handling policy. See here for valid
        values http://docs.python.org/2/library/codecs.html
    :returns: slugified unicode representation of `value`
    :raises TypeError: If text is not an instance of str
    """
    value = encodeutils.safe_decode(value, incoming, errors)
    # NOTE(aababilov): no need to use safe_(encode|decode) here:
    # encodings are always "ascii", error handling is always "ignore"
    # and types are always known (first: unicode; second: str)
    value = (
        unicodedata.normalize("NFKD", value)
        .encode("ascii", "ignore")
        .decode("ascii")
    )
    value = SLUGIFY_STRIP_RE.sub("", value).strip().lower()
    return SLUGIFY_HYPHENATE_RE.sub("-", value)


# NOTE(dhellmann): Before submitting a patch to add a new argument to
# this function to allow the caller to pass in "extra" or "additional"
# or "replacement" patterns to be masked out, please note that we have
# discussed that feature many times and always rejected it based on
# the desire to have Oslo functions behave consistently across all
# projects and *especially* to have security features work the same
# way no matter where they are used. If every project adopted its own
# set patterns for secret values, it would be very difficult to audit
# the logging to ensure that everything is properly masked. So, please
# either add your pattern to the module-level variables at the top of
# this file or, even better, pick an existing pattern or key to use in
# your application to ensure that the value is masked by this
# function.
def mask_password(message: str, secret: str = "***") -> str:  # noqa: S107
    """Replace password with *secret* in message.

    :param message: The string which includes security information.
    :param secret: value with which to replace passwords.
    :returns: The unicode value of message with the password fields masked.

    For example:

    >>> mask_password("'adminPass' : 'aaaaa'")
    "'adminPass' : '***'"
    >>> mask_password("'admin_pass' : 'aaaaa'")
    "'admin_pass' : '***'"
    >>> mask_password('"password" : "aaaaa"')
    '"password" : "***"'
    >>> mask_password("'original_password' : 'aaaaa'")
    "'original_password' : '***'"

    .. versionadded:: 0.2

    .. versionchanged:: 1.1
       Replace also ``'auth_token'``, ``'new_pass'`` and ``'auth_password'``
       keys.

    .. versionchanged:: 1.1.1
       Replace also ``'secret_uuid'`` key.

    .. versionchanged:: 1.5
       Replace also ``'sys_pswd'`` key.

    .. versionchanged:: 2.6
       Replace also ``'token'`` key.

    .. versionchanged:: 2.7
       Replace also ``'secret'`` key.

    .. versionchanged:: 3.4
       Replace also ``'configdrive'`` key.

    .. versionchanged:: 3.8
       Replace also ``'CHAPPASSWORD'`` key.
    """

    try:
        message = str(message)
    except UnicodeDecodeError:  # nosec
        # NOTE(jecarey): Temporary fix to handle cases where message is a
        # byte string. A better solution will be provided in Kilo.
        pass

    substitute1 = r'\g<1>' + secret
    substitute2 = r'\g<1>' + secret + r'\g<2>'
    substitute_wildcard = r'\g<1>'

    # NOTE(ldbragst): Check to see if anything in message contains any key
    # specified in _SANITIZE_KEYS, if not then just return the message since
    # we don't have to mask any passwords.
    for key in _SANITIZE_KEYS:
        if key in message.lower():
            for pattern in _SANITIZE_PATTERNS_2[key]:
                message = re.sub(pattern, substitute2, message)
            for pattern in _SANITIZE_PATTERNS_1[key]:
                message = re.sub(pattern, substitute1, message)
            # NOTE(hberaud): Those case are poorly handled by previous
            # patterns. They are passwords with quotes or double quotes.
            # They also needs a different way to substitute group this is why
            # they aren't fix in the pattern 1 or 2.
            for pattern in _SANITIZE_PATTERNS_WILDCARD[key]:
                message = re.sub(pattern, substitute_wildcard, message)
    return message


# TODO(stephenfin): The types aren't great for this. We want to indicate that
# the types of values in the returned dict are always identical to those of the
# input collection except if the value was a non-dict collection. It would be
# better if we returned the same type of Mapping here that we received on input.
def mask_dict_password(
    dictionary: collections.abc.Mapping[Any, Any],
    secret: str = "***",  # noqa: S107
) -> dict[str, Any]:
    """Replace password with *secret* in a dictionary recursively.

    :param dictionary: The dictionary which includes secret information.
    :param secret: value with which to replace secret information.
    :returns: The dictionary with string substitutions.

    A dictionary (which may contain nested dictionaries) contains
    information (such as passwords) which should not be revealed, and
    this function helps detect and replace those with the 'secret'
    provided (or `***` if none is provided).

    Substitution is performed in one of three situations:

    If the key is something that is considered to be indicative of a
    secret, then the corresponding value is replaced with the secret
    provided (or `***` if none is provided).

    If a value in the dictionary is a string, then it is masked
    using the ``mask_password()`` function.

    Finally, if a value is a dictionary, this function will
    recursively mask that dictionary as well.

    For example:

    >>> mask_dict_password({'password': 'd81juxmEW_',
    >>>                     'user': 'admin',
    >>>                     'home-dir': '/home/admin'},
    >>>                     '???')
    {'password': '???', 'user': 'admin', 'home-dir': '/home/admin'}

    For example (the value is masked using mask_password())

    >>> mask_dict_password({'password': '--password d81juxmEW_',
    >>>                     'user': 'admin',
    >>>                     'home-dir': '/home/admin'},
    >>>                     '???')
    {'password': '--password ???', 'user': 'admin', 'home-dir': '/home/admin'}

    For example (a nested dictionary is masked):

    >>> mask_dict_password({"nested": {'password': 'd81juxmEW_',
    >>>                     'user': 'admin',
    >>>                     'home': '/home/admin'}},
    >>>                     '???')
    {"nested": {'password': '???', 'user': 'admin', 'home': '/home/admin'}}

    .. versionadded:: 3.4

    """
    if not isinstance(dictionary, collections.abc.Mapping):
        raise TypeError(f"Expected a Mapping, got {type(dictionary)} instead.")

    out: dict[str, Any] = {}
    for k, v in dictionary.items():
        if isinstance(v, collections.abc.Mapping):
            out[k] = mask_dict_password(v, secret=secret)
            continue

        # NOTE(jlvillal): Check to see if anything in the dictionary 'key'
        # contains any key specified in _SANITIZE_KEYS.
        k_matched = False
        if isinstance(k, str):
            for sani_key in _SANITIZE_KEYS:
                if sani_key in k.lower():
                    out[k] = secret
                    k_matched = True
                    break
        if not k_matched:
            # We did not find a match for the key name in the
            # _SANITIZE_KEYS, so we fall through to here
            if isinstance(v, str):
                out[k] = mask_password(v, secret=secret)
            else:
                # Just leave it alone.
                out[k] = v

    return out


def is_int_like(val: Any) -> bool:
    """Check if a value looks like an integer with base 10.

    :param val: Value to verify
    :type val: string
    :returns: bool

    .. versionadded:: 1.1
    """
    try:
        return str(int(val)) == str(val)
    except (TypeError, ValueError):
        return False


def check_string_length(
    value: str,
    name: str | None = None,
    min_length: int = 0,
    max_length: int | None = None,
) -> None:
    """Check the length of specified string.

    :param value: the value of the string
    :param name: the name of the string
    :param min_length: the min_length of the string
    :param max_length: the max_length of the string
    :raises TypeError, ValueError: For any invalid input.

    .. versionadded:: 3.7
    """
    if name is None:
        name = value

    if not isinstance(value, str):
        msg = _("%s is not a string or unicode") % name
        raise TypeError(msg)

    length = len(value)
    if length < min_length:
        msg = _(
            "%(name)s has %(length)s characters, less than %(min_length)s."
        ) % {'name': name, 'length': length, 'min_length': min_length}
        raise ValueError(msg)

    if max_length and length > max_length:
        msg = _(
            "%(name)s has %(length)s characters, more than %(max_length)s."
        ) % {'name': name, 'length': length, 'max_length': max_length}
        raise ValueError(msg)


def validate_integer(
    value: int,
    name: str,
    min_value: int | None = None,
    max_value: int | None = None,
) -> int:
    """Make sure that value is a valid integer, potentially within range.

    :param value: value of the integer
    :param name: name of the integer
    :param min_value: min_value of the integer
    :param max_value: max_value of the integer
    :returns: integer
    :raises: ValueError if value is an invalid integer

    .. versionadded:: 3.33
    """
    try:
        value = int(str(value))
    except (ValueError, UnicodeEncodeError):
        msg = _('%(value_name)s must be an integer') % {'value_name': name}
        raise ValueError(msg)

    if min_value is not None and value < min_value:
        msg = _('%(value_name)s must be >= %(min_value)d') % {
            'value_name': name,
            'min_value': min_value,
        }
        raise ValueError(msg)

    if max_value is not None and value > max_value:
        msg = _('%(value_name)s must be <= %(max_value)d') % {
            'value_name': name,
            'max_value': max_value,
        }
        raise ValueError(msg)

    return value


def split_path(
    path: str,
    minsegs: int = 1,
    maxsegs: int | None = None,
    rest_with_last: bool = False,
) -> list[str | None]:
    """Validate and split the given HTTP request path.

    **Examples**::

        ['a'] = _split_path('/a')
        ['a', None] = _split_path('/a', 1, 2)
        ['a', 'c'] = _split_path('/a/c', 1, 2)
        ['a', 'c', 'o/r'] = _split_path('/a/c/o/r', 1, 3, True)

    :param path: HTTP Request path to be split
    :param minsegs: Minimum number of segments to be extracted
    :param maxsegs: Maximum number of segments to be extracted
    :param rest_with_last: If True, trailing data will be returned as part
                           of last segment.  If False, and there is
                           trailing data, raises ValueError.
    :returns: list of segments with a length of maxsegs (non-existent
              segments will return as None)
    :raises: ValueError if given an invalid path

    .. versionadded:: 3.11
    """
    if not maxsegs:
        maxsegs = minsegs
    if minsegs > maxsegs:
        raise ValueError(
            _('minsegs > maxsegs: %(min)d > %(max)d)')
            % {'min': minsegs, 'max': maxsegs}
        )

    if rest_with_last:
        segs = path.split('/', maxsegs)
        minsegs += 1
        maxsegs += 1
        count = len(segs)
        if (
            segs[0]
            or count < minsegs
            or count > maxsegs
            or '' in segs[1:minsegs]
        ):
            raise ValueError(_('Invalid path: %s') % urllib.parse.quote(path))
    else:
        minsegs += 1
        maxsegs += 1
        segs = path.split('/', maxsegs)
        count = len(segs)
        if (
            segs[0]
            or count < minsegs
            or count > maxsegs + 1
            or '' in segs[1:minsegs]
            or (count == maxsegs + 1 and segs[maxsegs])
        ):
            raise ValueError(_('Invalid path: %s') % urllib.parse.quote(path))

    segs = segs[1:maxsegs]
    return segs + [None] * (maxsegs - 1 - len(segs))


def split_by_commas(value: str) -> list[str]:
    """Split values by commas and quotes according to api-wg

    :param value: value to be split

    .. versionadded:: 3.17
    """
    # pyparsing is a slow import; defer loading until we need it
    import pyparsing as pp

    word = pp.QuotedString(quoteChar='"', escChar='\\') | pp.Word(
        pp.printables, excludeChars='",'
    )
    grammar = pp.stringStart + pp.delimitedList(word) + pp.stringEnd

    try:
        return list(grammar.parseString(value))
    except pp.ParseException:
        raise ValueError(f"Invalid value: {value}")



# Copyright 2014 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import sys
from typing import Any

import debtcollector.removals


def safe_decode(
    text: str | bytes,
    incoming: str | None = None,
    errors: str = 'strict',
) -> str:
    """Decodes incoming text/bytes string using `incoming` if they're not
       already unicode.

    :param incoming: Text's current encoding
    :param errors: Errors handling policy. See here for valid
        values http://docs.python.org/2/library/codecs.html
    :returns: text or a unicode `incoming` encoded
                representation of it.
    :raises TypeError: If text is not an instance of str
    """
    if not isinstance(text, str | bytes):
        raise TypeError(f"{type(text)} can't be decoded")

    if isinstance(text, str):
        return text

    if not incoming:
        incoming = (
            getattr(sys.stdin, 'encoding', None) or sys.getdefaultencoding()
        )

    try:
        return text.decode(incoming, errors)
    except UnicodeDecodeError:
        # Note(flaper87) If we get here, it means that
        # sys.stdin.encoding / sys.getdefaultencoding
        # didn't return a suitable encoding to decode
        # text. This happens mostly when global LANG
        # var is not set correctly and there's no
        # default encoding. In this case, most likely
        # python will use ASCII or ANSI encoders as
        # default encodings but they won't be capable
        # of decoding non-ASCII characters.
        #
        # Also, UTF-8 is being used since it's an ASCII
        # extension.
        return text.decode('utf-8', errors)


def safe_encode(
    text: str | bytes,
    incoming: str | None = None,
    encoding: str = 'utf-8',
    errors: str = 'strict',
) -> bytes:
    """Encodes incoming text/bytes string using `encoding`.

    If incoming is not specified, text is expected to be encoded with
    current python's default encoding. (`sys.getdefaultencoding`)

    :param incoming: Text's current encoding
    :param encoding: Expected encoding for text (Default UTF-8)
    :param errors: Errors handling policy. See here for valid
        values http://docs.python.org/2/library/codecs.html
    :returns: text or a bytestring `encoding` encoded
                representation of it.
    :raises TypeError: If text is not an instance of str

    See also to_utf8() function which is simpler and don't depend on
    the locale encoding.
    """
    if not isinstance(text, str | bytes):
        raise TypeError(f"{type(text)} can't be encoded")

    if not incoming:
        incoming = (
            getattr(sys.stdin, 'encoding', None) or sys.getdefaultencoding()
        )

    # Avoid case issues in comparisons
    if hasattr(incoming, 'lower'):
        incoming = incoming.lower()
    if hasattr(encoding, 'lower'):
        encoding = encoding.lower()

    if isinstance(text, str):
        return text.encode(encoding, errors)
    elif text and encoding != incoming:
        # Decode text before encoding it with `encoding`
        text = safe_decode(text, incoming, errors)
        return text.encode(encoding, errors)
    else:
        return text


def to_utf8(text: str | bytes) -> bytes:
    """Encode Unicode to UTF-8, return bytes unchanged.

    Raise TypeError if text is not a bytes string or a Unicode string.

    .. versionadded:: 3.5
    """
    if isinstance(text, bytes):
        return text
    elif isinstance(text, str):
        return text.encode('utf-8')
    else:
        raise TypeError(
            f"bytes or Unicode expected, got {type(text).__name__}"
        )


@debtcollector.removals.remove(
    message='Use str(exc) instead', category=DeprecationWarning
)
def exception_to_unicode(exc: Any) -> str:
    """Get the message of an exception as a Unicode string.

    On Python 3, the exception message is always a Unicode string.

    .. versionadded:: 1.6
    """
    return str(exc)



# Copyright (c) 2012 Intel Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
UUID related utilities and helper functions.

.. versionadded:: 1.1
"""

from typing import Any
import uuid


def generate_uuid(dashed: bool = True) -> str:
    """Creates a random uuid string.

    :param dashed: Generate uuid with dashes or not
    :type dashed: bool
    :returns: string
    """
    if dashed:
        return str(uuid.uuid4())
    return uuid.uuid4().hex


def _format_uuid_string(string: str) -> str:
    return (
        string.replace('urn:', '')
        .replace('uuid:', '')
        .strip('{}')
        .replace('-', '')
        .lower()
    )


def is_uuid_like(val: Any) -> bool:
    """Returns validation of a value as a UUID.

    :param val: Value to verify
    :type val: string
    :returns: bool

    .. versionchanged:: 1.1.1
       Support non-lowercase UUIDs.
    """
    try:
        return str(uuid.UUID(val)).replace('-', '') == _format_uuid_string(val)
    except (TypeError, ValueError, AttributeError):
        return False



# Copyright 2011 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Time related utilities and helper functions.
"""

from __future__ import annotations

import calendar
from collections.abc import Callable
import datetime
import functools
import logging
import time
from typing import Any
import zoneinfo

import iso8601

from oslo_utils import reflection

# ISO 8601 extended time format with microseconds
PERFECT_TIME_FORMAT = '%Y-%m-%dT%H:%M:%S.%f'

_MAX_DATETIME_SEC = 59

now = time.monotonic


def parse_isotime(timestr: str) -> datetime.datetime:
    """Parse time from ISO 8601 format."""
    try:
        return iso8601.parse_date(timestr)
    except iso8601.ParseError as e:
        raise ValueError(str(e))
    except TypeError as e:
        raise ValueError(str(e))


def parse_strtime(
    timestr: str, fmt: str = PERFECT_TIME_FORMAT
) -> datetime.datetime:
    """Turn a formatted time back into a datetime."""
    return datetime.datetime.strptime(timestr, fmt)


def normalize_time(timestamp: datetime.datetime) -> datetime.datetime:
    """Normalize time in arbitrary timezone to UTC naive object."""
    offset = timestamp.utcoffset()
    if offset is None:
        return timestamp
    return timestamp.replace(tzinfo=None) - offset


# https://github.com/python/mypy/issues/2087
class _UTCNow:
    override_time: None | datetime.datetime | list[datetime.datetime] = None

    def __call__(self, with_timezone: bool = False) -> datetime.datetime:
        """Overridable version of utils.utcnow that can return a TZ-aware datetime.

        See :py:class:`oslo_utils.fixture.TimeFixture`.

        .. versionchanged:: 1.6
           Added *with_timezone* parameter.
        """
        if self.override_time:
            if isinstance(self.override_time, datetime.datetime):
                return self.override_time
            return self.override_time.pop(0)
        if with_timezone:
            return datetime.datetime.now(tz=iso8601.iso8601.UTC)
        return datetime.datetime.now(datetime.timezone.utc).replace(
            tzinfo=None
        )


utcnow = _UTCNow()


def is_older_than(
    before: str | datetime.datetime, seconds: int | float
) -> bool:
    """Return True if before is older than seconds.

    .. versionchanged:: 1.7
       Accept datetime string with timezone information.
       Fix comparison with timezone aware datetime.
    """
    if isinstance(before, str):
        before = parse_isotime(before)

    before = normalize_time(before)

    return utcnow() - before > datetime.timedelta(seconds=seconds)


def is_newer_than(
    after: str | datetime.datetime, seconds: int | float
) -> bool:
    """Return True if after is newer than seconds.

    .. versionchanged:: 1.7
       Accept datetime string with timezone information.
       Fix comparison with timezone aware datetime.
    """
    if isinstance(after, str):
        after = parse_isotime(after)

    after = normalize_time(after)

    return after - utcnow() > datetime.timedelta(seconds=seconds)


def utcnow_ts(microsecond: bool = False) -> float:
    """Timestamp version of our utcnow function.

    See :py:class:`oslo_utils.fixture.TimeFixture`.

    .. versionchanged:: 1.3
       Added optional *microsecond* parameter.
    """
    if utcnow.override_time is None:
        # NOTE(kgriffs): This is several times faster
        # than going through calendar.timegm(...)
        timestamp = time.time()
        if not microsecond:
            timestamp = int(timestamp)
        return timestamp

    now = utcnow()
    timestamp = calendar.timegm(now.timetuple())

    if microsecond:
        timestamp += float(now.microsecond) / 1000000

    return timestamp


def set_time_override(override_time: datetime.datetime | None = None) -> None:
    """Overrides utils.utcnow.

    Make it return a constant time or a list thereof, one at a time.

    See :py:class:`oslo_utils.fixture.TimeFixture`.

    :param override_time: datetime instance or list thereof. If not
                          given, defaults to the current UTC time.
    """
    utcnow.override_time = override_time or datetime.datetime.now(
        datetime.timezone.utc
    ).replace(tzinfo=None)


def advance_time_delta(timedelta: datetime.timedelta) -> None:
    """Advance overridden time using a datetime.timedelta.

    See :py:class:`oslo_utils.fixture.TimeFixture`.

    """
    if utcnow.override_time is None:
        raise RuntimeError('override_time must be configured')

    if isinstance(utcnow.override_time, datetime.datetime):
        utcnow.override_time += timedelta
    else:
        for dt in utcnow.override_time:
            dt += timedelta


def advance_time_seconds(seconds: int | float) -> None:
    """Advance overridden time by seconds.

    See :py:class:`oslo_utils.fixture.TimeFixture`.

    """
    advance_time_delta(datetime.timedelta(0, seconds))


def clear_time_override() -> None:
    """Remove the overridden time.

    See :py:class:`oslo_utils.fixture.TimeFixture`.

    """
    utcnow.override_time = None


def marshall_now(now: datetime.datetime | None = None) -> dict[str, Any]:
    """Make an rpc-safe datetime with microseconds.

    .. versionchanged:: 1.6
       Timezone information is now serialized instead of being stripped.
    """
    if not now:
        now = utcnow()
    d: dict[str, str | int | None] = dict(
        day=now.day,
        month=now.month,
        year=now.year,
        hour=now.hour,
        minute=now.minute,
        second=now.second,
        microsecond=now.microsecond,
    )
    if now.tzinfo:
        # Need to handle either iso8601 or python UTC format
        tzname = now.tzinfo.tzname(None)
        d['tzname'] = 'UTC' if tzname == 'UTC+00:00' else tzname
    return d


def unmarshall_time(tyme: dict[str, Any]) -> datetime.datetime:
    """Unmarshall a datetime dict.

    .. versionchanged:: 1.5
       Drop leap second.

    .. versionchanged:: 1.6
       Added support for timezone information.
    """

    # NOTE(ihrachys): datetime does not support leap seconds,
    # so the best thing we can do for now is dropping them
    # http://bugs.python.org/issue23574
    second = min(tyme['second'], _MAX_DATETIME_SEC)
    dt = datetime.datetime(
        day=tyme['day'],
        month=tyme['month'],
        year=tyme['year'],
        hour=tyme['hour'],
        minute=tyme['minute'],
        second=second,
        microsecond=tyme['microsecond'],
    )
    tzname = tyme.get('tzname')
    if tzname:
        # Need to handle either iso8601 or python UTC format
        tzname = 'UTC' if tzname == 'UTC+00:00' else tzname
        tzinfo = zoneinfo.ZoneInfo(tzname)
        dt = dt.replace(tzinfo=tzinfo)

    return dt


def delta_seconds(
    before: datetime.datetime, after: datetime.datetime
) -> float:
    """Return the difference between two timing objects.

    Compute the difference in seconds between two date, time, or
    datetime objects (as a float, to microsecond resolution).
    """
    delta = after - before
    return delta.total_seconds()


def is_soon(dt: datetime.datetime, window: int | float) -> bool:
    """Determines if time is going to happen in the next window seconds.

    :param dt: the time
    :param window: minimum seconds to remain to consider the time not soon

    :return: True if expiration is within the given duration
    """
    soon = utcnow() + datetime.timedelta(seconds=window)
    return normalize_time(dt) <= soon


class Split:
    """A *immutable* stopwatch split.

    See: http://en.wikipedia.org/wiki/Stopwatch for what this is/represents.

    .. versionadded:: 1.4
    """

    __slots__ = ['_elapsed', '_length']

    def __init__(self, elapsed: float, length: float) -> None:
        self._elapsed = elapsed
        self._length = length

    @property
    def elapsed(self) -> float:
        """Duration from stopwatch start."""
        return self._elapsed

    @property
    def length(self) -> float:
        """Seconds from last split (or the elapsed time if no prior split)."""
        return self._length

    def __repr__(self) -> str:
        r = reflection.get_class_name(self, fully_qualified=False)
        r += f"(elapsed={self._elapsed}, length={self._length})"
        return r


def time_it(
    logger: logging.Logger,
    log_level: int = logging.DEBUG,
    message: str = "It took %(seconds).02f seconds to run function '%(func_name)s'",
    enabled: bool = True,
    min_duration: float = 0.01,
) -> Callable[[Callable[..., Any]], Callable[..., Any]]:
    """Decorator that will log how long its decorated function takes to run.

    This does **not** output a log if the decorated function fails
    with an exception.

    :param logger: logger instance to use when logging elapsed time
    :param log_level: logger logging level to use when logging elapsed time
    :param message: customized message to use when logging elapsed time,
                    the message may use automatically provide values
                    ``%(seconds)`` and ``%(func_name)`` if it finds those
                    values useful to record
    :param enabled: whether to enable or disable this decorator (useful to
                    decorate a function with this decorator, and then easily
                    be able to switch that decoration off by some config or
                    other value)
    :param min_duration: argument that determines if logging is triggered
                         or not, it is by default set to 0.01 seconds to avoid
                         logging when durations and/or elapsed function call
                         times are less than 0.01 seconds, to disable
                         any ``min_duration`` checks this value should be set
                         to less than or equal to zero or set to none
    """

    def decorator(func: Callable[..., Any]) -> Callable[..., Any]:
        if not enabled:
            return func

        @functools.wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            with StopWatch() as w:
                result = func(*args, **kwargs)
            time_taken = w.elapsed()
            if min_duration is None or time_taken >= min_duration:
                logger.log(
                    log_level,
                    message,
                    {
                        'seconds': time_taken,
                        'func_name': reflection.get_callable_name(func),
                    },
                )
            return result

        return wrapper

    return decorator


class StopWatch:
    """A simple timer/stopwatch helper class.

    Inspired by: apache-commons-lang java stopwatch.

    Not thread-safe (when a single watch is mutated by multiple threads at
    the same time). Thread-safe when used by a single thread (not shared) or
    when operations are performed in a thread-safe manner on these objects by
    wrapping those operations with locks.

    .. versionadded:: 1.4
    """

    _STARTED = 'STARTED'
    _STOPPED = 'STOPPED'

    def __init__(self, duration: float | None = None) -> None:
        if duration is not None and duration < 0:
            raise ValueError(
                f"Duration must be greater or equal to zero and not {duration}"
            )
        self._duration = duration
        self._started_at: float | None = None
        self._stopped_at: float | None = None
        self._state: str | None = None
        self._splits: tuple[Split, ...] = ()

    def start(self) -> StopWatch:
        """Starts the watch (if not already started).

        NOTE(harlowja): resets any splits previously captured (if any).
        """
        if self._state == self._STARTED:
            return self
        self._started_at = now()
        self._stopped_at = None
        self._state = self._STARTED
        self._splits = ()
        return self

    @property
    def splits(self) -> tuple[Split, ...]:
        """Accessor to all/any splits that have been captured."""
        return self._splits

    def split(self) -> Split:
        """Captures a split/elapsed since start time (and doesn't stop)."""
        if self._state == self._STARTED:
            elapsed = self.elapsed()
            if self._splits:
                length = self._delta_seconds(self._splits[-1].elapsed, elapsed)
            else:
                length = elapsed
            self._splits = self._splits + (Split(elapsed, length),)
            return self._splits[-1]
        else:
            raise RuntimeError(
                "Can not create a split time of a stopwatch"
                " if it has not been started or if it has been"
                " stopped"
            )

    def restart(self) -> StopWatch:
        """Restarts the watch from a started/stopped state."""
        if self._state == self._STARTED:
            self.stop()
        self.start()
        return self

    @staticmethod
    def _delta_seconds(earlier: float, later: float) -> float:
        # Uses max to avoid the delta/time going backwards (and thus negative).
        return max(0.0, later - earlier)

    def elapsed(self, maximum: float | None = None) -> float:
        """Returns how many seconds have elapsed."""
        if (
            self._state not in (self._STARTED, self._STOPPED)
            or self._started_at is None
        ):
            raise RuntimeError(
                "Can not get the elapsed time of a stopwatch "
                "if it has not been started/stopped"
            )

        if self._state == self._STOPPED:
            if self._stopped_at is None:
                raise RuntimeError("internal error")

            elapsed = self._delta_seconds(self._started_at, self._stopped_at)
        else:
            elapsed = self._delta_seconds(self._started_at, now())
        if maximum is not None and elapsed > maximum:
            elapsed = max(0.0, maximum)
        return elapsed

    def __enter__(self) -> StopWatch:
        """Starts the watch."""
        self.start()
        return self

    def __exit__(self, type: Any, value: Any, traceback: Any) -> None:
        """Stops the watch (ignoring errors if stop fails)."""
        try:
            self.stop()
        except RuntimeError:  # nosec: errors are meant to be ignored
            pass

    def leftover(self, return_none: bool = False) -> float | None:
        """Returns how many seconds are left until the watch expires.

        :param return_none: when ``True`` instead of raising a ``RuntimeError``
                            when no duration has been set this call will
                            return ``None`` instead.
        :type return_none: boolean
        """
        if self._state != self._STARTED:
            raise RuntimeError(
                "Can not get the leftover time of a stopwatch"
                " that has not been started"
            )
        if self._duration is None:
            if not return_none:
                raise RuntimeError(
                    "Can not get the leftover time of a watch"
                    " that has no duration"
                )
            return None
        return max(0.0, self._duration - self.elapsed())

    def expired(self) -> bool:
        """Returns if the watch has expired (ie, duration provided elapsed)."""
        if self._state not in (self._STARTED, self._STOPPED):
            raise RuntimeError(
                "Can not check if a stopwatch has expired"
                " if it has not been started/stopped"
            )
        if self._duration is None:
            return False
        return self.elapsed() > self._duration

    def has_started(self) -> bool:
        """Returns True if the watch is in a started state."""
        return self._state == self._STARTED

    def has_stopped(self) -> bool:
        """Returns True if the watch is in a stopped state."""
        return self._state == self._STOPPED

    def resume(self) -> None:
        """Resumes the watch from a stopped state."""
        if self._state == self._STOPPED:
            self._state = self._STARTED
        else:
            raise RuntimeError(
                "Can not resume a stopwatch that has not been stopped"
            )

    def stop(self) -> StopWatch | None:
        """Stops the watch."""
        if self._state == self._STOPPED:
            return None
        if self._state != self._STARTED:
            raise RuntimeError(
                "Can not stop a stopwatch that has not been started"
            )
        self._stopped_at = now()
        self._state = self._STOPPED
        return self



# Copyright 2012, Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import threading
from unittest import mock
import warnings

import eventlet
from eventlet import greenthread
from oslotest import base as test_base

from oslo_utils import eventletutils


class EventletUtilsTest(test_base.BaseTestCase):
    def setUp(self):
        super().setUp()
        self._old_avail = eventletutils.EVENTLET_AVAILABLE
        eventletutils.EVENTLET_AVAILABLE = True

    def tearDown(self):
        super().tearDown()
        eventletutils.EVENTLET_AVAILABLE = self._old_avail

    @mock.patch("oslo_utils.eventletutils._patcher")
    def test_warning_not_patched(self, mock_patcher):
        mock_patcher.already_patched = True
        mock_patcher.is_monkey_patched.return_value = False
        with warnings.catch_warnings(record=True) as capture:
            warnings.simplefilter("always")
            eventletutils.warn_eventlet_not_patched(['os'])
        self.assertEqual(1, len(capture))
        w = capture[0]
        self.assertEqual(RuntimeWarning, w.category)
        self.assertIn('os', str(w.message))

    @mock.patch("oslo_utils.eventletutils._patcher")
    def test_warning_not_patched_none_provided(self, mock_patcher):
        mock_patcher.already_patched = True
        mock_patcher.is_monkey_patched.return_value = False
        with warnings.catch_warnings(record=True) as capture:
            warnings.simplefilter("always")
            eventletutils.warn_eventlet_not_patched()
        self.assertEqual(1, len(capture))
        w = capture[0]
        self.assertEqual(RuntimeWarning, w.category)
        for m in eventletutils._ALL_PATCH:
            self.assertIn(m, str(w.message))

    @mock.patch("oslo_utils.eventletutils._patcher")
    def test_warning_not_patched_all(self, mock_patcher):
        mock_patcher.already_patched = True
        mock_patcher.is_monkey_patched.return_value = False
        with warnings.catch_warnings(record=True) as capture:
            warnings.simplefilter("always")
            eventletutils.warn_eventlet_not_patched(['all'])
        self.assertEqual(1, len(capture))
        w = capture[0]
        self.assertEqual(RuntimeWarning, w.category)
        for m in eventletutils._ALL_PATCH:
            self.assertIn(m, str(w.message))

    @mock.patch("oslo_utils.eventletutils._patcher")
    def test_no_warning(self, mock_patcher):
        mock_patcher.already_patched = True
        mock_patcher.is_monkey_patched.return_value = True
        with warnings.catch_warnings(record=True) as capture:
            warnings.simplefilter("always")
            eventletutils.warn_eventlet_not_patched(['os'])
        self.assertEqual(0, len(capture))

    @mock.patch("oslo_utils.eventletutils._patcher")
    def test_eventlet_is_patched(self, mock_patcher):
        mock_patcher.is_monkey_patched.return_value = True
        self.assertTrue(eventletutils.is_monkey_patched('os'))
        mock_patcher.is_monkey_patched.return_value = False
        self.assertFalse(eventletutils.is_monkey_patched('os'))

    @mock.patch("oslo_utils.eventletutils._patcher", None)
    def test_eventlet_no_patcher(self):
        self.assertFalse(eventletutils.is_monkey_patched('os'))

    @mock.patch("oslo_utils.eventletutils._patcher")
    def test_partially_patched_warning(self, mock_patcher):
        is_patched = set()
        mock_patcher.already_patched = True
        mock_patcher.is_monkey_patched.side_effect = lambda m: m in is_patched
        with warnings.catch_warnings(record=True) as capture:
            warnings.simplefilter("always")
            eventletutils.warn_eventlet_not_patched(['os'])
        self.assertEqual(1, len(capture))
        is_patched.add('os')
        with warnings.catch_warnings(record=True) as capture:
            warnings.simplefilter("always")
            eventletutils.warn_eventlet_not_patched(['os'])
        self.assertEqual(0, len(capture))
        is_patched.add('thread')
        with warnings.catch_warnings(record=True) as capture:
            warnings.simplefilter("always")
            eventletutils.warn_eventlet_not_patched(['os', 'thread'])
        self.assertEqual(0, len(capture))
        with warnings.catch_warnings(record=True) as capture:
            warnings.simplefilter("always")
            eventletutils.warn_eventlet_not_patched(['all'])
        self.assertEqual(1, len(capture))
        w = capture[0]
        self.assertEqual(RuntimeWarning, w.category)
        for m in ['os', 'thread']:
            self.assertNotIn(m, str(w.message))

    def test_invalid_patch_check(self):
        self.assertRaises(
            ValueError, eventletutils.warn_eventlet_not_patched, ['blah.blah']
        )

    @mock.patch('oslo_utils.eventletutils._eventlet')
    def test_event_api_compat(self, mock_eventlet):
        with mock.patch(
            'oslo_utils.eventletutils.is_monkey_patched', return_value=True
        ):
            e_event = eventletutils.Event()
        self.assertIsInstance(e_event, eventletutils.EventletEvent)

        t_event = eventletutils.Event()
        t_event_cls = threading.Event
        self.assertIsInstance(t_event, t_event_cls)

        public_methods = [
            m
            for m in dir(t_event)
            if not m.startswith("_") and callable(getattr(t_event, m))
        ]

        for method in public_methods:
            self.assertTrue(hasattr(e_event, method))

        # Ensure set() allows multiple invocations, same as in
        # threading implementation.
        e_event.set()
        self.assertTrue(e_event.isSet())
        e_event.set()
        self.assertTrue(e_event.isSet())

    def test_event_no_timeout(self):
        event = eventletutils.EventletEvent()

        def thread_a():
            self.assertTrue(event.wait())

        a = greenthread.spawn(thread_a)

        with eventlet.timeout.Timeout(0.5, False):
            a.wait()
            self.fail('wait() timed out')

    def test_event_race(self):
        event = eventletutils.EventletEvent()

        def thread_a():
            self.assertTrue(event.wait(2))

        a = greenthread.spawn(thread_a)

        def thread_b():
            eventlet.sleep(0.1)
            event.clear()
            event.set()
            a.wait()

        b = greenthread.spawn(thread_b)
        with eventlet.timeout.Timeout(0.5):
            b.wait()

    def test_event_clear_timeout(self):
        event = eventletutils.EventletEvent()

        def thread_a():
            self.assertFalse(event.wait(0.5))

        a = greenthread.spawn(thread_a)

        def thread_b():
            eventlet.sleep(0.1)
            event.clear()
            eventlet.sleep(0.1)
            event.clear()
            a.wait()

        b = greenthread.spawn(thread_b)
        with eventlet.timeout.Timeout(0.7):
            b.wait()

    def test_event_set_clear_timeout(self):
        event = eventletutils.EventletEvent()
        wakes = []

        def thread_func():
            result = event.wait(0.2)
            wakes.append(result)
            if len(wakes) == 1:
                self.assertTrue(result)
                event.clear()
            else:
                self.assertFalse(result)

        a = greenthread.spawn(thread_func)
        b = greenthread.spawn(thread_func)
        eventlet.sleep(0)  # start threads
        event.set()

        with eventlet.timeout.Timeout(0.3):
            a.wait()
            b.wait()
        self.assertFalse(event.is_set())
        self.assertEqual([True, False], wakes)

    @mock.patch('oslo_utils.eventletutils._eventlet.event.Event')
    def test_event_clear_already_sent(self, mock_event):
        old_event = mock.Mock()
        new_event = mock.Mock()
        mock_event.side_effect = [old_event, new_event]
        event = eventletutils.EventletEvent()
        event.set()
        event.clear()
        self.assertEqual(1, old_event.send.call_count)



# Copyright 2010 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All Rights Reserved.
# Copyright (c) 2010 Citrix Systems, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Helper methods to deal with images.

.. versionadded:: 3.1

.. versionchanged:: 3.14.0
   add paramter format.

"""

import json
import re
from typing import Any

import debtcollector

from oslo_utils._i18n import _
from oslo_utils import strutils


class QemuImgInfo:
    """Parse Qemu image information from command `qemu-img info`'s output.

    The instance of :class:`QemuImgInfo` has properties: `image`,
    `backing_file`, `file_format`, `virtual_size`, `cluster_size`,
    `disk_size`, `snapshots` and `encrypted`.

    The parameter format can be set to 'json' or 'human'. With 'json' format
    output, qemu image information will be parsed more easily and readable.
    However 'human' format support will be dropped in next cycle and only
    'json' format will be supported. Prefer to use 'json' instead of 'human'.
    """

    BACKING_FILE_RE = re.compile(
        (
            r"^(.*?)\s*\(actual\s+path\s*:"
            r"\s+(.*?)\)\s*$"
        ),
        re.I,
    )
    TOP_LEVEL_RE = re.compile(r"^([\w\d\s\_\-]+):(.*)$")
    SIZE_RE = re.compile(
        r"([0-9]+[eE][-+][0-9]+|\d*\.?\d+)"
        r"\s*(\w+)?(\s*\(\s*(\d+)\s+bytes\s*\))?",
        re.I,
    )

    def __init__(
        self,
        cmd_output: str | bytes | bytearray | None = None,
        format: str = 'human',
    ) -> None:
        if isinstance(cmd_output, bytes | bytearray):
            cmd_output = cmd_output.decode()

        if format == 'json':
            details = json.loads(cmd_output or '{}')
            self.image = details.get('filename')
            self.backing_file = details.get('backing-filename')
            self.backing_file_format = details.get('backing-filename-format')
            self.file_format = details.get('format')
            self.virtual_size = details.get('virtual-size')
            self.cluster_size = details.get('cluster-size')
            self.disk_size = details.get('actual-size')
            self.snapshots = details.get('snapshots', [])
            self.encrypted = 'yes' if details.get('encrypted') else None
            self.format_specific = details.get('format-specific')
        else:
            if cmd_output is not None:
                debtcollector.deprecate(
                    'The human format is deprecated and the format parameter '
                    'will be removed. Use explicitly json instead',
                    version="xena",
                    category=FutureWarning,
                )
            details = self._parse(cmd_output or '')
            self.image = details.get('image')
            self.backing_file = details.get('backing_file')
            self.backing_file_format = details.get('backing_file_format')
            self.file_format = details.get('file_format')
            self.virtual_size = details.get('virtual_size')
            self.cluster_size = details.get('cluster_size')
            self.disk_size = details.get('disk_size')
            self.snapshots = details.get('snapshot_list', [])
            self.encrypted = details.get('encrypted')
            self.format_specific = None

    def __str__(self) -> str:
        lines = [
            f'image: {self.image}',
            f'file_format: {self.file_format}',
            f'virtual_size: {self.virtual_size}',
            f'disk_size: {self.disk_size}',
            f'cluster_size: {self.cluster_size}',
            f'backing_file: {self.backing_file}',
            f'backing_file_format: {self.backing_file_format}',
        ]
        if self.snapshots:
            lines.append(f"snapshots: {self.snapshots}")
        if self.encrypted:
            lines.append(f"encrypted: {self.encrypted}")
        if self.format_specific:
            lines.append(f"format_specific: {self.format_specific}")
        return "\n".join(lines)

    def _canonicalize(self, field: str) -> str:
        # Standardize on underscores/lc/no dash and no spaces
        # since qemu seems to have mixed outputs here... and
        # this format allows for better integration with python
        # - i.e. for usage in kwargs and such...
        field = field.lower().strip()
        for c in (" ", "-"):
            field = field.replace(c, '_')
        return field

    def _extract_bytes(self, details: str) -> int:
        # Replace it with the byte amount
        real_size = self.SIZE_RE.search(details)
        if not real_size:
            raise ValueError(_('Invalid input value "%s".') % details)
        magnitude = real_size.group(1)
        if "e" in magnitude.lower():
            magnitude = format(float(real_size.group(1)), '.0f')
        unit_of_measure = real_size.group(2)
        bytes_info = real_size.group(3)
        if bytes_info:
            return int(real_size.group(4))
        elif not unit_of_measure:
            return int(magnitude)
        # Allow abbreviated unit such as K to mean KB for compatibility.
        if len(unit_of_measure) == 1 and unit_of_measure != 'B':
            unit_of_measure += 'B'
        return int(
            strutils.string_to_bytes(
                f'{magnitude}{unit_of_measure}', return_int=True
            )
        )

    def _extract_details(
        self, root_cmd: str, root_details: str, lines_after: list[str]
    ) -> str | int | list[dict[str, str]]:
        real_details: str | int | list[dict[str, str]] = root_details
        if root_cmd == 'backing_file':
            # Replace it with the real backing file
            backing_match = self.BACKING_FILE_RE.match(root_details)
            if backing_match:
                real_details = backing_match.group(2).strip()
        elif root_cmd in ['virtual_size', 'cluster_size', 'disk_size']:
            # Replace it with the byte amount (if we can convert it)
            if root_details in ('None', 'unavailable'):
                real_details = 0
            else:
                real_details = self._extract_bytes(root_details)
        elif root_cmd == 'file_format':
            real_details = root_details.strip().lower()
        elif root_cmd == 'snapshot_list':
            # Next line should be a header, starting with 'ID'
            if not lines_after or not lines_after.pop(0).startswith("ID"):
                msg = _("Snapshot list encountered but no header found!")
                raise ValueError(msg)
            real_details = []
            # This is the sprintf pattern we will try to match
            # "%-10s%-20s%7s%20s%15s"
            # ID TAG VM SIZE DATE VM CLOCK (current header)
            while lines_after:
                line = lines_after[0]
                line_pieces = line.split()
                if len(line_pieces) != 6:
                    break
                # Check against this pattern in the final position
                # "%02d:%02d:%02d.%03d"
                date_pieces = line_pieces[5].split(":")
                if len(date_pieces) != 3:
                    break
                lines_after.pop(0)
                real_details.append(
                    {
                        'id': line_pieces[0],
                        'tag': line_pieces[1],
                        'vm_size': line_pieces[2],
                        'date': line_pieces[3],
                        'vm_clock': line_pieces[4] + " " + line_pieces[5],
                    }
                )
        return real_details

    def _parse(self, cmd_output: str) -> dict[str, Any]:
        # Analysis done of qemu-img.c to figure out what is going on here
        # Find all points start with some chars and then a ':' then a newline
        # and then handle the results of those 'top level' items in a separate
        # function.
        #
        # TODO(harlowja): newer versions might have a json output format
        #                 we should switch to that whenever possible.
        #                 see: http://bit.ly/XLJXDX
        contents = {}
        lines = [x for x in cmd_output.splitlines() if x.strip()]
        while lines:
            line = lines.pop(0)
            top_level = self.TOP_LEVEL_RE.match(line)
            if top_level:
                root = self._canonicalize(top_level.group(1))
                if not root:
                    continue
                root_details = top_level.group(2).strip()
                details = self._extract_details(root, root_details, lines)
                contents[root] = details
        return contents



# Copyright (c) 2011 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ast
from collections.abc import Callable
import operator
from typing import Any

import pyparsing


def _all_in(x: str, *y: str) -> bool:
    x = ast.literal_eval(x)
    if not isinstance(x, list):
        raise TypeError(
            "<all-in> must compare with a list literal"
            " string, EG \"{}\"".format(['aes', 'mmx'])
        )
    return all(val in x for val in y)


def _range_in(x: str, *y: str) -> bool:
    x = ast.literal_eval(x)
    if len(y) != 4:
        raise TypeError(
            "<range-in> operator has to be followed by 2 "
            "space separated numeric value surrounded by "
            "brackets \"range_in [ 10 20 ] \""
        )
    num_x = float(x)
    num_y = float(y[1])
    num_z = float(y[2])
    if num_y > num_z:
        raise TypeError(
            "<range-in> operator's first argument has to be "
            "smaller or equal to the second argument EG"
            "\"range_in  ( 10 20 ] \""
        )

    if y[0] == '[':
        lower = num_x >= num_y
    elif y[0] == '(':
        lower = num_x > num_y
    else:
        raise TypeError(
            "The first element should be an opening bracket (\"(\" or \"[\")"
        )

    if y[3] == ']':
        upper = num_x <= num_z
    elif y[3] == ')':
        upper = num_x < num_z
    else:
        raise TypeError(
            "The last element should be a closing bracket (\")\" or \"]\")"
        )
    return lower and upper


op_methods: dict[str, Callable[..., bool]] = {
    # This one is special/odd,
    # TODO(harlowja): fix it so that it's not greater than or
    # equal, see here for the original @ https://review.openstack.org/#/c/8089/
    '=': lambda x, y: float(x) >= float(y),
    # More sane ops/methods
    # Numerical methods
    '!=': lambda x, y: float(x) != float(y),
    '<=': lambda x, y: float(x) <= float(y),
    '<': lambda x, y: float(x) < float(y),
    '==': lambda x, y: float(x) == float(y),
    '>=': lambda x, y: float(x) >= float(y),
    '>': lambda x, y: float(x) > float(y),
    # String methods
    's!=': operator.ne,
    's<': operator.lt,
    's<=': operator.le,
    's==': operator.eq,
    's>': operator.gt,
    's>=': operator.ge,
    # Other
    '<all-in>': _all_in,
    '<in>': lambda x, y: y in x,
    '<or>': lambda x, *y: any(x == a for a in y),
    '<range-in>': _range_in,
}


def make_grammar() -> pyparsing.ParserElement:
    """Creates the grammar to be used by a spec matcher.

    The grammar created supports the following operations.

    Numerical values:
      * ``=  :`` equal to or greater than. This is equivalent to ``>=`` and is
        supported for `legacy reasons
        <http://docs.openstack.org/developer/nova/filter_scheduler.html#ComputeCapabilitiesFilter>`_
      * ``!= :`` Float/integer value not equal
      * ``<= :`` Float/integer value less than or equal
      * ``<  :`` Float/integer value less than
      * ``== :`` Float/integer value equal
      * ``>= :`` Float/integer value greater than or equal
      * ``>  :`` Float/integer value greater

    String operations:
      * ``s!= :`` Not equal
      * ``s<  :`` Less than
      * ``s<= :`` Less than or equal
      * ``s== :`` Equal
      * ``s>  :`` Greater than
      * ``s>= :`` Greater than or equal

    Other operations:
      * ``<all-in>  :`` All items 'in' value
      * ``<in>      :`` Item 'in' value, like a substring in a string.
      * ``<or>      :`` Logical 'or'
      * ``<range-in>:`` Range tester with customizable boundary conditions, tests
                        whether value is in the range, boundary condition could be
                        inclusve \'[\' or exclusive \'(\'.

    If no operator is specified the default is ``s==`` (string equality comparison)

    Example operations:
     * ``">= 60"`` Is the numerical value greater than or equal to 60
     * ``"<or> spam <or> eggs"`` Does the value contain ``spam`` or ``eggs``
     * ``"s== 2.1.0"`` Is the string value equal to ``2.1.0``
     * ``"<in> gcc"`` Is the string ``gcc`` contained in the value string
     * ``"<all-in> aes mmx"`` Are both ``aes`` and ``mmx`` in the value
     * ``"<range-in> [ 10 20 ]"`` float(value) >= 10 and float(value) <= 20
     * ``"<range-in> ( 10 20 ]"`` float(value) > 10 and float(value) <= 20
     * ``"<range-in> ( 10 20 )"`` float(value) > 10 and float(value) < 20

    :returns: A pyparsing.MatchFirst object. See
              https://pythonhosted.org/pyparsing/ for details on pyparsing.
    """
    # This is apparently how pyparsing recommends to be used,
    # as http://pyparsing.wikispaces.com/share/view/644825 states that
    # it is not thread-safe to use a parser across threads.

    unary_ops = (
        # Order matters here (so that '=' doesn't match before '==')
        pyparsing.Literal("==")
        | pyparsing.Literal("=")
        | pyparsing.Literal("!=")
        | pyparsing.Literal("<in>")
        | pyparsing.Literal(">=")
        | pyparsing.Literal("<=")
        | pyparsing.Literal(">")
        | pyparsing.Literal("<")
        | pyparsing.Literal("s==")
        | pyparsing.Literal("s!=")
        |
        # Order matters here (so that '<' doesn't match before '<=')
        pyparsing.Literal("s<=")
        | pyparsing.Literal("s<")
        |
        # Order matters here (so that '>' doesn't match before '>=')
        pyparsing.Literal("s>=")
        | pyparsing.Literal("s>")
    )

    all_in_nary_op = pyparsing.Literal("<all-in>")
    or_ = pyparsing.Literal("<or>")
    range_in_binary_op = pyparsing.Literal("<range-in>")

    # An atom is anything not an keyword followed by anything but whitespace
    atom = ~(
        unary_ops | all_in_nary_op | or_ | range_in_binary_op
    ) + pyparsing.Regex(r"\S+")

    unary = unary_ops + atom
    range_op = range_in_binary_op + atom + atom + atom + atom
    nary = all_in_nary_op + pyparsing.OneOrMore(atom)
    disjunction = pyparsing.OneOrMore(or_ + atom)

    # Even-numbered tokens will be '<or>', so we drop them
    disjunction.setParseAction(lambda _s, _l, t: ["<or>"] + t[1::2])

    expr = disjunction | nary | range_op | unary | atom
    return expr


def match(cmp_value: Any, spec: str) -> bool:
    """Match a given value to a given spec DSL.

    This uses the grammar defined by make_grammar()

    :param cmp_value: Value to be checked for match.
    :param spec: The comparison specification string, for example ``">= 70"``
                 or ``"s== string_value"``. See ``make_grammar()`` for examples
                 of a specification string.
    :returns: True if cmp_value is a match for spec. False otherwise.
    """
    expr = make_grammar()
    tree: pyparsing.ParseResults | list[str]
    try:
        # As of 2018-01-29 documentation on parseString()
        # https://pythonhosted.org/pyparsing/pyparsing.ParserElement-class.html#parseString
        #
        # parseString() will take our specification string, for example "< 6"
        # and convert it into a list of ['<', "6"]
        tree = expr.parseString(spec)
    except pyparsing.ParseException:
        # If an exception then we will just check if the value matches the spec
        tree = [spec]
    if len(tree) == 1:
        return bool(tree[0] == cmp_value)

    # tree[0] will contain a string representation of a comparison operation
    # such as '>=', we then convert that string to a comparison function
    compare_func = op_methods[tree[0]]
    return compare_func(cmp_value, *tree[1:])



# Copyright (c) 2016 EasyStack Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from oslotest import base as test_base

from oslo_utils import dictutils as du


class DictUtilsTestCase(test_base.BaseTestCase):
    def test_flatten_dict_to_keypairs(self):
        data = {'a': 'A', 'b': 'B', 'nested': {'a': 'A', 'b': 'B'}}
        pairs = list(du.flatten_dict_to_keypairs(data))
        self.assertEqual(
            [('a', 'A'), ('b', 'B'), ('nested:a', 'A'), ('nested:b', 'B')],
            pairs,
        )

    def test_flatten_dict_to_keypairs_with_separator(self):
        data = {'a': 'A', 'b': 'B', 'nested': {'a': 'A', 'b': 'B'}}
        pairs = list(du.flatten_dict_to_keypairs(data, separator='.'))
        self.assertEqual(
            [('a', 'A'), ('b', 'B'), ('nested.a', 'A'), ('nested.b', 'B')],
            pairs,
        )
